import numpy as np
import scipy.spatial.qhull as qhull
from functools import wraps
from sklearn.cluster import KMeans

from autolens import decorator_util
from autolens import exc
from autolens.array import mask as msk, scaled_array
from autolens.array.util import grid_util, array_util, mask_util, binning_util
from autolens.array.mapping_util import (
    array_mapping_util,
    grid_mapping_util,
    sparse_mapping_util,
)


def check_input_grid_and_options_are_compatible(grid):

    if not isinstance(grid, Grid):
        raise exc.GridException(
            "You are trying to return an array from a _from_grid function that is mapped to 2d or a binned up "
            "sub grid. However, the input grid is not an instance of the Grid class. You must make the"
            "input grid a Grid."
        )


def reshape_array(func):
    @wraps(func)
    def wrapper(object, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 1D arrays of physical quantities \
        (e.g. image, convergences, potentials). Depending on the input variables, it determines whether the
        returned array is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        object : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        grid : ndarray or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.

        Returns
        -------
            An array of a physical quantity that may be in 1D or 2D and binned up from a sub-grid.
        """

        return_in_2d = kwargs["return_in_2d"] if "return_in_2d" in kwargs else False

        grid = object.grid

        result = func(object)

        if len(result.shape) == 2:
            result_1d = grid.sub_array_1d_from_sub_array_2d(sub_array_2d=result)
        else:
            result_1d = result

        if not return_in_2d:
            return result_1d
        else:
            return grid.scaled_array_2d_from_array_1d(array_1d=result_1d)

    return wrapper


def reshape_array_from_grid_and_psf(func):
    @wraps(func)
    def wrapper(object, grid, psf, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 1D arrays of physical quantities \
        (e.g. image, convergences, potentials). Depending on the input variables, it determines whether the
        returned array is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        object : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        grid : ndarray or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        Returns
        -------
            An array of a physical quantity that may be in 1D or 2D and binned up from a sub-grid.
        """

        return_in_2d = kwargs["return_in_2d"] if "return_in_2d" in kwargs else True
        result = func(object, grid, psf)

        return reshape_array_result(grid=grid, result=result, return_in_2d=return_in_2d)

    return wrapper


def reshape_array_from_grid_and_convolver(func):
    @wraps(func)
    def wrapper(object, grid, convolver, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 1D arrays of physical quantities \
        (e.g. image, convergences, potentials). Depending on the input variables, it determines whether the
        returned array is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        object : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        grid : ndarray or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        Returns
        -------
            An array of a physical quantity that may be in 1D or 2D and binned up from a sub-grid.
        """

        return_in_2d = kwargs["return_in_2d"] if "return_in_2d" in kwargs else True
        result = func(object, grid, convolver)

        return reshape_array_result(grid=grid, result=result, return_in_2d=return_in_2d)

    return wrapper


def reshape_array_from_grid(func):
    @wraps(func)
    def wrapper(object, grid, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 1D arrays of physical quantities \
        (e.g. image, convergences, potentials). Depending on the input variables, it determines whether the
        returned array is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        object : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        grid : ndarray or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.

        Returns
        -------
            An array of a physical quantity that may be in 1D or 2D and binned up from a sub-grid.
        """

        return_in_2d = kwargs["return_in_2d"] if "return_in_2d" in kwargs else True
        return_binned = kwargs["return_binned"] if "return_binned" in kwargs else True
        result = func(object, grid)

        return reshape_sub_array_result(
            grid=grid,
            result=result,
            return_in_2d=return_in_2d,
            return_binned=return_binned,
        )

    return wrapper


def reshape_array_result(grid, result, return_in_2d):

    if len(result.shape) == 2:
        result_1d = grid.array_1d_from_array_2d(array_2d=result)
    else:
        result_1d = result

    if not return_in_2d:
        return result_1d
    else:
        return grid.scaled_array_2d_from_array_1d(array_1d=result_1d)


def reshape_sub_array_result(grid, result, return_in_2d, return_binned):

    if len(result.shape) == 2:
        result_1d = grid.sub_array_1d_from_sub_array_2d(sub_array_2d=result)
    else:
        result_1d = result

    if not return_in_2d and not return_binned:
        return result_1d

    check_input_grid_and_options_are_compatible(grid=grid)

    if not return_in_2d and not return_binned:

        return result_1d

    elif not return_in_2d and return_binned:

        return grid.array_1d_binned_from_sub_array_1d(sub_array_1d=result_1d)

    elif return_in_2d and not return_binned:

        return grid.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
            sub_array_1d=result_1d
        )

    elif return_in_2d and return_binned:

        return grid.scaled_array_2d_binned_from_sub_array_1d(sub_array_1d=result_1d)


def reshape_returned_grid(func):
    @wraps(func)
    def wrapper(object, grid, *args, **kwargs):
        """

        This wrapper decorates the _from_grid functions of profiles, which return 2D grids of physical quantities \
        (e.g. deflection angles). Depending on the input variables, it determines whether the
        returned grid is reshaped to 2D from 1D and if a sub-grid is input, it can bin the sub-gridded values to
        gridded values.

        Parameters
        ----------
        object : autolens.model.geometry_profiles.Profile
            The profiles that owns the function
        grid : ndgrid or Grid or Grid
            (y,x) in either cartesian or profiles coordinate system

        Returns
        -------
            An grid of (y,x) coordinates that may be in 1D or 2D and binned up from a sub-grid.
        """

        return_in_2d = kwargs["return_in_2d"] if "return_in_2d" in kwargs else False
        return_binned = kwargs["return_binned"] if "return_binned" in kwargs else False

        result = func(object, grid)

        if len(result.shape) == 3:
            result_1d = grid.sub_grid_1d_with_sub_dimensions_from_sub_grid_2d(
                sub_grid_2d=result
            )
        else:
            result_1d = result

        if not return_in_2d and not return_binned:
            return result_1d

        check_input_grid_and_options_are_compatible(grid=grid)

        if not return_in_2d and not return_binned:

            return result_1d

        elif not return_in_2d and return_binned:

            return grid.grid_1d_binned_from_sub_grid_1d(sub_grid_1d=result_1d)

        elif return_in_2d and not return_binned:

            return grid.sub_grid_2d_with_sub_dimensions_from_sub_grid_1d(
                sub_grid_1d=result_1d
            )

        elif return_in_2d and return_binned:

            return grid.grid_2d_binned_from_sub_grid_1d(sub_grid_1d=result_1d)

    return wrapper


class Grid(np.ndarray):
    def __new__(
        cls,
        arr,
        mask,
        sub_grid_size=1,
        border_filter_size=None,
        binned=None,
        *args,
        **kwargs
    ):
        """A grid of coordinates, where each entry corresponds to the (y,x) coordinates at the centre of an \
        unmasked pixel. The positive y-axis is upwards and poitive x-axis to the right. 
        
        A *Grid* is ordered such pixels begin from the top-row of the mask and go rightwards and then \ 
        downwards. Therefore, it is a ndarray of shape [total_unmasked_pixels, 2]. The first element of the ndarray \
        thus corresponds to the pixel index and second element the y or x arc -econd coordinates. For example:

        - grid[3,0] = the 4th unmasked pixel's y-coordinate.
        - grid[6,1] = the 7th unmasked pixel's x-coordinate.

        Below is a visual illustration of a grid, where a total of 10 pixels are unmasked and are included in \
        the grid.

        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     This is an example mask.Mask, where:
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|o|o|x|x|x|x|     x = True (Pixel is masked and excluded from the grid)
        |x|x|x|o|o|o|o|x|x|x|     o = False (Pixel is not masked and included in the grid)
        |x|x|x|o|o|o|o|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|

        The mask pixel index's will come out like this (and the direction of arc-second coordinates is highlighted
        around the mask.

        pixel_scale = 1.0"

        <--- -ve  x  +ve -->
                                                        y      x
        |x|x|x|x|x|x|x|x|x|x|  ^   grid[0] = [ 1.5, -0.5]
        |x|x|x|x|x|x|x|x|x|x|  |   grid[1] = [ 1.5,  0.5]
        |x|x|x|x|x|x|x|x|x|x|  |   grid[2] = [ 0.5, -1.5]
        |x|x|x|x|0|1|x|x|x|x| +ve  grid[3] = [ 0.5, -0.5]
        |x|x|x|2|3|4|5|x|x|x|  y   grid[4] = [ 0.5,  0.5]
        |x|x|x|6|7|8|9|x|x|x| -ve  grid[5] = [ 0.5,  1.5]
        |x|x|x|x|x|x|x|x|x|x|  |   grid[6] = [-0.5, -1.5]
        |x|x|x|x|x|x|x|x|x|x|  |   grid[7] = [-0.5, -0.5]
        |x|x|x|x|x|x|x|x|x|x| \/   grid[8] = [-0.5,  0.5]
        |x|x|x|x|x|x|x|x|x|x|      grid[9] = [-0.5,  1.5]

        A sub-grid of coordinates, where each entry corresponds to the (y,x) coordinates at the centre of each \
        sub-pixel of an unmasked pixel (e.g. the pixels of a grid). The positive y-axis is upwards and poitive \
        x-axis to the right, and this convention is followed for the sub-pixels in each unmasked pixel.

        A *SubGrid* is ordered such that pixels begin from the first (top-left) sub-pixel in the first unmasked pixel. \
        Indexes then go over the sub-pixels in each unmasked pixel, for every unmasked pixel. Therefore, \
        the sub-grid is an ndarray of shape [total_unmasked_pixels*(sub_grid_shape)**2, 2]. For example:

        - grid[9, 1] - using a 2x2 sub-grid, gives the 3rd unmasked pixel's 2nd sub-pixel x-coordinate.
        - grid[9, 1] - using a 3x3 sub-grid, gives the 2nd unmasked pixel's 1st sub-pixel x-coordinate.
        - grid[27, 0] - using a 3x3 sub-grid, gives the 4th unmasked pixel's 1st sub-pixel y-coordinate.

        Below is a visual illustration of a sub grid. Like the grid, the indexing of each sub-pixel goes from \
        the top-left corner. In contrast to the grid above, our illustration below restricts the mask to just \
        2 pixels, to keep the illustration brief.

        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     This is an example mask.Mask, where:
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     x = True (Pixel is masked and excluded from lens)
        |x|x|x|x|o|o|x|x|x|x|     o = False (Pixel is not masked and included in lens)
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|

        Our grid looks like it did before:

        pixel_scale = 1.0"

        <--- -ve  x  +ve -->

        |x|x|x|x|x|x|x|x|x|x|  ^
        |x|x|x|x|x|x|x|x|x|x|  |
        |x|x|x|x|x|x|x|x|x|x|  |                        y     x
        |x|x|x|x|x|x|x|x|x|x| +ve  grid[0] = [0.5,  -1.5]
        |x|x|x|0|1|x|x|x|x|x|  y   grid[1] = [0.5,  -0.5]
        |x|x|x|x|x|x|x|x|x|x| -ve
        |x|x|x|x|x|x|x|x|x|x|  |
        |x|x|x|x|x|x|x|x|x|x|  |
        |x|x|x|x|x|x|x|x|x|x| \/
        |x|x|x|x|x|x|x|x|x|x|

        However, we now go to each unmasked pixel and derive a sub-pixel grid for it. For example, for pixel 0,
        if *sub_grid_size=2*, we use a 2x2 sub-grid:

        Pixel 0 - (2x2):
                                y      x
               grid[0] = [0.66, -1.66]
        |0|1|  grid[1] = [0.66, -1.33]
        |2|3|  grid[2] = [0.33, -1.66]
               grid[3] = [0.33, -1.33]

        Now, we'd normally sub-grid all pixels using the same *sub_grid_size*, but for this illustration lets
        pretend we used a sub_grid_size of 3x3 for pixel 1:

                                  y      x
                 grid[0] = [0.75, -0.75]
                 grid[1] = [0.75, -0.5]
                 grid[2] = [0.75, -0.25]
        |0|1|2|  grid[3] = [0.5,  -0.75]
        |3|4|5|  grid[4] = [0.5,  -0.5]
        |6|7|8|  grid[5] = [0.5,  -0.25]
                 grid[6] = [0.25, -0.75]
                 grid[7] = [0.25, -0.5]
                 grid[8] = [0.25, -0.25]

        """
        obj = arr.view(cls)
        obj.mask = mask
        obj.sub_grid_size = sub_grid_size
        obj.sub_grid_length = int(sub_grid_size ** 2.0)
        obj.sub_grid_fraction = 1.0 / obj.sub_grid_length
        obj.border_filter_size = border_filter_size
        obj.sub_border_pixels = mask.sub_border_1d_indexes_from_sub_grid_size(
            sub_grid_size=sub_grid_size, filter_size=border_filter_size
        )
        obj.interpolator = None
        obj.binned = None
        return obj

    def __array_finalize__(self, obj):

        if isinstance(obj, Grid):
            self.mask = obj.mask
            self.sub_grid_size = obj.sub_grid_size
            self.sub_grid_length = obj.sub_grid_length
            self.sub_grid_fraction = obj.sub_grid_fraction
            self.border_filter_size = obj.border_filter_size
            self.sub_border_pixels = obj.sub_border_pixels
            self.interpolator = obj.interpolator
            self.binned = obj.binned

    @classmethod
    def from_mask_and_sub_grid_size(
        cls, mask, sub_grid_size=1, border_filter_size=None
    ):
        """Setup a sub-grid of the unmasked pixels, using a mask and a specified sub-grid size. The center of \
        every unmasked pixel's sub-pixels give the grid's (y,x) arc-second coordinates.

        Parameters
        -----------
        mask : Mask
            The mask whose masked pixels are used to setup the sub-pixel grid.
        sub_grid_size : int
            The size (sub_grid_size x sub_grid_size) of each unmasked pixels sub-grid.
        """

        sub_grid_masked = grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
            mask=mask, pixel_scales=mask.pixel_scales, sub_grid_size=sub_grid_size
        )

        return Grid(
            arr=sub_grid_masked,
            mask=mask,
            sub_grid_size=sub_grid_size,
            border_filter_size=border_filter_size,
        )

    @classmethod
    def from_shape_pixel_scale_and_sub_grid_size(
        cls, shape, pixel_scale, sub_grid_size=1, border_filter_size=None
    ):
        """Setup a sub-grid from a 2D array shape and pixel scale. Here, the center of every pixel on the 2D \
        array gives the grid's (y,x) arc-second coordinates, where each pixel has sub-pixels specified by the \
        sub-grid size.

        This is equivalent to using a 2D mask consisting entirely of unmasked pixels.

        Parameters
        -----------
        shape : (int, int)
            The 2D shape of the array, where all pixels are used to generate the grid's grid.
        pixel_scale : float
            The size of each pixel in arc seconds.
        sub_grid_size : int
            The size (sub_grid_size x sub_grid_size) of each unmasked pixels sub-grid.
        """

        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(
            shape=shape, pixel_scale=pixel_scale
        )

        grid = grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
            mask=mask, pixel_scales=mask.pixel_scales, sub_grid_size=sub_grid_size
        )

        return Grid(
            arr=grid,
            mask=mask,
            sub_grid_size=sub_grid_size,
            border_filter_size=border_filter_size,
        )

    @classmethod
    def blurring_grid_from_mask_and_psf_shape(cls, mask, psf_shape):
        """Setup a blurring-grid from a mask, where a blurring grid consists of all pixels that are masked, but they \
        are close enough to the unmasked pixels that a fraction of their light will be blurred into those pixels \
        via PSF convolution. For example, if our mask is as follows:
        
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     This is an ccd.Mask, where:
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|     x = True (Pixel is masked and excluded from lens)
        |x|x|x|o|o|o|x|x|x|x|     o = False (Pixel is not masked and included in lens)
        |x|x|x|o|o|o|x|x|x|x|
        |x|x|x|o|o|o|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|x|
        
        For a PSF of shape (3,3), the following blurring mask is computed (noting that only pixels that are direct \
        neighbors of the unmasked pixels above will blur light into an unmasked pixel):

        |x|x|x|x|x|x|x|x|x|     This is an example grid.Mask, where:
        |x|x|x|x|x|x|x|x|x|
        |x|x|o|o|o|o|o|x|x|     x = True (Pixel is masked and excluded from lens)
        |x|x|o|x|x|x|o|x|x|     o = False (Pixel is not masked and included in lens)
        |x|x|o|x|x|x|o|x|x|
        |x|x|o|x|x|x|o|x|x|
        |x|x|o|o|o|o|o|x|x|
        |x|x|x|x|x|x|x|x|x|
        |x|x|x|x|x|x|x|x|x|
        
        Thus, the blurring grid coordinates and indexes will be as follows:
        
        pixel_scale = 1.0"

        <--- -ve  x  +ve -->
                                                            y     x
        |x|x|x |x |x |x |x |x|x|  |   blurring_grid[0] = [2.0, -2.0]  blurring_grid[9] =  [-1.0, -2.0]
        |x|x|x |x |x |x |x |x|x|  |   blurring_grid[1] = [2.0, -1.0]  blurring_grid[10] = [-1.0,  2.0]
        |x|x|0 |1 |2 |3 |4 |x|x| +ve  blurring_grid[2] = [2.0,  0.0]  blurring_grid[11] = [-2.0, -2.0]
        |x|x|5 |x |x |x |6 |x|x|  y   blurring_grid[3] = [2.0,  1.0]  blurring_grid[12] = [-2.0, -1.0]
        |x|x|7 |x |x |x |8 |x|x| -ve  blurring_grid[4] = [2.0,  2.0]  blurring_grid[13] = [-2.0,  0.0]
        |x|x|9 |x |x |x |10|x|x|  |   blurring_grid[5] = [1.0, -2.0]  blurring_grid[14] = [-2.0,  1.0]
        |x|x|11|12|13|14|15|x|x|  |   blurring_grid[6] = [1.0,  2.0]  blurring_grid[15] = [-2.0,  2.0]
        |x|x|x |x |x |x |x |x|x| \/   blurring_grid[7] = [0.0, -2.0]
        |x|x|x |x |x |x |x |x|x|      blurring_grid[8] = [0.0,  2.0]
        
        For a PSF of shape (5,5), the following blurring mask is computed (noting that pixels that are 2 pixels from an
        direct unmasked pixels now blur light into an unmasked pixel):

        |x|x|x|x|x|x|x|x|x|     This is an example grid.Mask, where:
        |x|o|o|o|o|o|o|o|x|
        |x|o|o|o|o|o|o|o|x|     x = True (Pixel is masked and excluded from lens)
        |x|o|o|x|x|x|o|o|x|     o = False (Pixel is not masked and included in lens)
        |x|o|o|x|x|x|o|o|x|
        |x|o|o|x|x|x|o|o|x|
        |x|o|o|o|o|o|o|o|x|
        |x|o|o|o|o|o|o|o|x|
        |x|x|x|x|x|x|x|x|x|
        """

        blurring_mask = mask.blurring_mask_from_psf_shape(psf_shape=psf_shape)

        return Grid.from_mask_and_sub_grid_size(mask=blurring_mask, sub_grid_size=1)

    @property
    def in_2d(self):
        return self.mask.sub_grid_2d_with_sub_dimensions_from_sub_grid_1d_and_sub_grid_size(
            sub_grid_1d=self, sub_grid_size=self.sub_grid_size
        )

    @property
    @array_util.Memoizer()
    def unlensed_unsubbed_1d(self):
        return Grid(
            arr=grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
                mask=self.mask, pixel_scales=self.mask.pixel_scales, sub_grid_size=1
            ),
            mask=self.mask,
            sub_grid_size=1,
            border_filter_size=self.border_filter_size,
        )

    @property
    def unlensed_1d(self):
        return Grid(
            arr=grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
                mask=self.mask,
                pixel_scales=self.mask.pixel_scales,
                sub_grid_size=self.sub_grid_size,
            ),
            mask=self.mask,
            sub_grid_size=self.sub_grid_size,
            border_filter_size=self.border_filter_size,
        )

    @property
    def unlensed_unmasked_1d(self):
        return Grid(
            grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
                mask=np.full(self.mask.shape, False),
                pixel_scales=self.mask.pixel_scales,
                sub_grid_size=self.sub_grid_size,
            ),
            mask=self.mask,
            sub_grid_size=self.sub_grid_size,
            border_filter_size=self.border_filter_size,
        )

    @property
    def unlensed_2d(self):
        return Grid(
            arr=grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
                sub_grid_1d=self.unlensed_1d,
                mask=self.mask,
                sub_grid_size=self.sub_grid_size,
            ),
            mask=self.mask,
            sub_grid_size=self.sub_grid_size,
            border_filter_size=self.border_filter_size,
        )

    @property
    def unlensed_unmasked_2d(self):
        return Grid(
            arr=grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
                mask=np.full(self.mask.shape, False),
                sub_grid_1d=self.unlensed_unmasked_1d,
                sub_grid_size=self.sub_grid_size,
            ),
            mask=self.mask,
            sub_grid_size=self.sub_grid_size,
            border_filter_size=self.border_filter_size,
        )

    @property
    def total_pixels(self):
        return self.shape[0]

    @property
    def pixel_scale(self):
        return self.mask.pixel_scale

    @property
    def sub_mask(self):

        sub_shape = (
            self.mask.shape[0] * self.sub_grid_size,
            self.mask.shape[1] * self.sub_grid_size,
        )

        sub_one_to_two = self.mask.sub_mask_1d_index_to_sub_mask_2d_index_from_sub_grid_size(
            sub_grid_size=self.sub_grid_size
        )

        return mask_util.mask_from_shape_and_mask_1d_index_to_mask_2d_index(
            shape=sub_shape, mask_1d_index_to_mask_2d_index=sub_one_to_two
        )

    def blurring_grid_from_psf_shape(self, psf_shape):

        blurring_mask = self.mask.blurring_mask_from_psf_shape(psf_shape=psf_shape)

        return Grid.from_mask_and_sub_grid_size(mask=blurring_mask, sub_grid_size=1)

    def marching_squares_grid_pixels_to_grid_arcsec(self, grid_pixels, shape):

        grid_arcsec = grid_util.grid_pixels_1d_to_grid_arcsec_1d(
            grid_pixels_1d=grid_pixels,
            shape=shape,
            pixel_scales=(
                self.pixel_scale / self.sub_grid_size,
                self.pixel_scale / self.sub_grid_size,
            ),
            origin=self.mask.origin,
        )

        grid_arcsec[:, 0] -= self.pixel_scale / (2.0 * self.sub_grid_size)
        grid_arcsec[:, 1] += self.pixel_scale / (2.0 * self.sub_grid_size)

        return grid_arcsec

    def array_2d_from_array_1d(self, array_1d):
        """ Map a 1D array the same dimension as the grid to its original 2D array. 
        
        Values which were masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array which is mapped to its masked 2D array.
        """
        return self.mask.array_2d_from_array_1d(array_1d=array_1d)

    def scaled_array_2d_from_array_1d(self, array_1d):
        """ Map a 1D array the same dimension as the grid to its original masked 2D array and return it as a hyper \
        array.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array of which is mapped to a 2D hyper array.
        """
        return self.mask.scaled_array_2d_from_array_1d(array_1d=array_1d)

    def array_1d_from_array_2d(self, array_2d):
        """ Map a 2D array to its masked 1D array..

        Values which are masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array which is mapped to its masked 2D array.
        """
        return self.mask.array_1d_from_array_2d(array_2d=array_2d)

    def grid_2d_from_grid_1d(self, grid_1d):
        """Map a 1D grid the same dimension as the grid to its original 2D grid.

        Values which were masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        grid_1d : ndarray
            The 1D grid which is mapped to its masked 2D array.
        """
        return self.mask.grid_2d_from_grid_1d(grid_1d=grid_1d)

    def grid_1d_from_grid_2d(self, grid_2d):
        """ Map a 2D grid to its masked 1D grid.. 

        Values which are masked in the mapping_util to the 1D grid are returned as zeros.

        Parameters
        -----------
        grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D grid.
        """
        return self.mask.grid_1d_from_grid_2d(grid_2d=grid_2d)

    def sub_array_2d_from_sub_array_1d(self, sub_array_1d):
        """ Map a 1D sub-array the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub_array which is mapped to its masked 2D sub-array.
        """
        return self.mask.sub_array_2d_from_sub_array_1d_and_sub_grid_size(
            sub_array_1d=sub_array_1d, sub_grid_size=self.sub_grid_size
        )

    def scaled_array_2d_with_sub_dimensions_from_sub_array_1d(self, sub_array_1d):
        """ Map a 1D sub-array the same dimension as the sub-grid to its original masked 2D sub-array and return it as
        a hyper array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub-array of which is mapped to a 2D hyper sub-array the dimensions.
        """
        return self.mask.scaled_array_2d_with_sub_dimensions_from_sub_array_1d_and_sub_grid_size(
            sub_array_1d=sub_array_1d, sub_grid_size=self.sub_grid_size
        )

    def scaled_array_2d_binned_from_sub_array_1d(self, sub_array_1d):
        """ Map a 1D sub-array the same dimension as the sub-grid to its original masked 2D sub-array and return it as
        a hyper array.

        Parameters
        -----------
        sub_array_1d : ndarray
            The 1D sub-array of which is mapped to a 2D hyper sub-array the dimensions.
        """
        return self.mask.scaled_array_2d_binned_from_sub_array_1d_and_sub_grid_size(
            sub_array_1d=sub_array_1d, sub_grid_size=self.sub_grid_size
        )

    def array_1d_binned_from_sub_array_1d(self, sub_array_1d):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_array_1d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """
        return self.mask.array_1d_binned_from_sub_array_1d_and_sub_grid_size(
            sub_array_1d=sub_array_1d, sub_grid_size=self.sub_grid_size
        )

    def sub_array_1d_from_sub_array_2d(self, sub_array_2d):
        """ Map a 2D sub-array to its masked 1D sub-array.

        Values which are masked in the mapping_util to the 1D array are returned as zeros.

        Parameters
        -----------
        su_array_2d : ndarray
            The 2D sub-array which is mapped to its masked 1D sub-array.
        """
        return self.mask.sub_array_1d_with_sub_dimensions_from_sub_array_2d_and_sub_grid_size(
            sub_array_2d=sub_array_2d, sub_grid_size=self.sub_grid_size
        )

    def grid_1d_binned_from_sub_grid_1d(self, sub_grid_1d):
        """ Map a 1D sub-grid the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub grid.

        Parameters
        -----------
        sub_grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D sub-grid.
        """
        return self.mask.grid_1d_binned_from_sub_grid_1d_and_sub_grid_size(
            sub_grid_1d=sub_grid_1d, sub_grid_size=self.sub_grid_size
        )

    def grid_2d_binned_from_sub_grid_1d(self, sub_grid_1d):
        """ Map a 1D sub-grid the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub grid.

        Parameters
        -----------
        sub_grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D sub-grid.
        """
        return self.mask.grid_2d_binned_from_sub_grid_1d_and_sub_grid_size(
            sub_grid_1d=sub_grid_1d, sub_grid_size=self.sub_grid_size
        )

    def sub_grid_1d_with_sub_dimensions_from_sub_grid_2d(self, sub_grid_2d):
        """For an input 1D sub-array, map its values to a 1D array of values by summing each set \of sub-pixel \
        values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_grid_2d : ndarray
            A 1D sub-array of values (e.g. image, convergence, potential) which is mapped to
            a 1d array.
        """
        return self.mask.sub_grid_1d_with_sub_dimensions_from_sub_grid_2d_and_sub_grid_size(
            sub_grid_2d=sub_grid_2d, sub_grid_size=self.sub_grid_size
        )

    def sub_grid_2d_with_sub_dimensions_from_sub_grid_1d(self, sub_grid_1d):
        """ Map a 1D sub-grid the same dimension as the sub-grid (e.g. including sub-pixels) to its original masked
        2D sub grid.

        Parameters
        -----------
        sub_grid_1d : ndgrid
            The 1D grid which is mapped to its masked 2D sub-grid.
        """
        return self.mask.sub_grid_2d_with_sub_dimensions_from_sub_grid_1d_and_sub_grid_size(
            sub_grid_1d=sub_grid_1d, sub_grid_size=self.sub_grid_size
        )

    def new_grid_with_binned_grid(self, binned_grid):
        # noinspection PyAttributeOutsideInit
        # TODO: This function doesn't do what it says on the tin. The returned grid would be the same as the grid
        # TODO: on which the function was called but with a new interpolator set.
        self.binned = binned_grid
        return self

    def new_grid_with_interpolator(self, pixel_scale_interpolation_grid):
        # noinspection PyAttributeOutsideInit
        # TODO: This function doesn't do what it says on the tin. The returned grid would be the same as the grid
        # TODO: on which the function was called but with a new interpolator set.
        self.interpolator = Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=self.mask,
            grid=self[:, :],
            pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        )
        return self

    def padded_grid_from_psf_shape(self, psf_shape):

        shape = self.mask.shape

        padded_shape = (shape[0] + psf_shape[0] - 1, shape[1] + psf_shape[1] - 1)

        padded_mask = msk.Mask.unmasked_for_shape_and_pixel_scale(
            shape=padded_shape, pixel_scale=self.mask.pixel_scale
        )

        padded_sub_grid = Grid.from_mask_and_sub_grid_size(
            mask=padded_mask, sub_grid_size=self.sub_grid_size
        )

        if self.interpolator is None:
            return padded_sub_grid
        else:
            return padded_sub_grid.new_grid_with_interpolator(
                pixel_scale_interpolation_grid=self.interpolator.pixel_scale_interpolation_grid
            )

    def trimmed_array_2d_from_padded_array_1d_and_image_shape(
        self, padded_array_1d, image_shape
    ):
        """ Map a padded 1D array of values to its original 2D array, trimming all edge values.

        Parameters
        -----------
        padded_array_1d : ndarray
            A 1D array of values which were computed using a padded grid
        """

        # TODO : Raise error if padded grid not used here.

        padded_array_2d = self.array_2d_from_array_1d(array_1d=padded_array_1d)
        pad_size_0 = self.mask.shape[0] - image_shape[0]
        pad_size_1 = self.mask.shape[1] - image_shape[1]
        return padded_array_2d[
            pad_size_0 // 2 : self.mask.shape[0] - pad_size_0 // 2,
            pad_size_1 // 2 : self.mask.shape[1] - pad_size_1 // 2,
        ]

    def convolve_array_1d_with_psf(self, padded_array_1d, psf):
        """Convolve a 1d padded array of values (e.g. image before PSF blurring) with a PSF, and then trim \
        the convolved array to its original 2D shape.

        Parameters
        -----------
        padded_array_1d: ndarray
            A 1D array of values which were computed using the a padded grid.
        psf : ndarray
            An array describing the PSF kernel of the image.
        """

        padded_array_2d = array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=padded_array_1d,
            mask=np.full(fill_value=False, shape=self.mask.shape),
            sub_grid_size=1,
        )

        # noinspection PyUnresolvedReferences
        blurred_padded_array_2d = psf.convolve(array_2d=padded_array_2d)

        return array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=blurred_padded_array_2d,
            mask=np.full(self.mask.shape, False),
            sub_grid_size=1,
        )

    def unmasked_blurred_array_2d_from_padded_array_1d_psf_and_image_shape(
        self, padded_array_1d, psf, image_shape
    ):
        """For a padded grid and psf, compute an unmasked blurred image from an unmasked unblurred image.

        This relies on using the lens data's padded-grid, which is a grid of (y,x) coordinates which extends over the \
        entire image as opposed to just the masked region.

        Parameters
        ----------
        psf : abstract_data.PSF
            The PSF of the image used for convolution.
        unmasked_image_1d : ndarray
            The 1D unmasked image which is blurred.
        """

        padded_array_2d = self.scaled_array_2d_from_array_1d(array_1d=padded_array_1d)

        blurred_image_2d = psf.convolve(array_2d=padded_array_2d)

        blurred_image_1d = self.array_1d_from_array_2d(array_2d=blurred_image_2d)

        return self.trimmed_array_2d_from_padded_array_1d_and_image_shape(
            padded_array_1d=blurred_image_1d, image_shape=image_shape
        )

    @property
    def sub_border_grid(self):
        return self[self.sub_border_pixels]

    def relocated_grid_from_grid(self, grid):
        """Determine a set of relocated grid from an input set of grid, by relocating their pixels based on the \
        borders.

        The blurring-grid does not have its coordinates relocated, as it is only used for computing analytic \
        light-profiles and not inversion-grid.

        Parameters
        -----------
        grid : GridStack
            The grid, whose grid coordinates are relocated.
        """

        return Grid(
            arr=self.relocated_grid_from_grid_jit(
                grid=grid, border_grid=self.sub_border_grid
            ),
            mask=grid.mask,
            sub_grid_size=grid.sub_grid_size,
            border_filter_size=self.border_filter_size,
        )

    def relocated_pixelization_grid_from_pixelization_grid(self, pixelization_grid):
        """Determine a set of relocated grid from an input set of grid, by relocating their pixels based on the \
        borders.

        The blurring-grid does not have its coordinates relocated, as it is only used for computing analytic \
        light-profiles and not inversion-grid.

        Parameters
        -----------
        grid : GridStack
            The grid, whose grid coordinates are relocated.
        """

        return PixelizationGrid(
            arr=self.relocated_grid_from_grid_jit(
                grid=pixelization_grid, border_grid=self.sub_border_grid
            ),
            mask_1d_index_to_nearest_pixelization_1d_index=pixelization_grid.mask_1d_index_to_nearest_pixelization_1d_index,
        )

    @staticmethod
    @decorator_util.jit()
    def relocated_grid_from_grid_jit(grid, border_grid):
        """ Relocate the coordinates of a grid to its border if they are outside the border. This is performed as \
        follows:

        1) Use the mean value of the grid's y and x coordinates to determine the origin of the grid.
        2) Compute the radial distance of every grid coordinate from the origin.
        3) For every coordinate, find its nearest pixel in the border.
        4) Determine if it is outside the border, by comparing its radial distance from the origin to its paid \
           border pixel's radial distance.
        5) If its radial distance is larger, use the ratio of radial distances to move the coordinate to the border \
           (if its inside the border, do nothing).
        """

        border_origin = np.zeros(2)
        border_origin[0] = np.mean(border_grid[:, 0])
        border_origin[1] = np.mean(border_grid[:, 1])
        border_grid_radii = np.sqrt(
            np.add(
                np.square(np.subtract(border_grid[:, 0], border_origin[0])),
                np.square(np.subtract(border_grid[:, 1], border_origin[1])),
            )
        )
        border_min_radii = np.min(border_grid_radii)

        grid_radii = np.sqrt(
            np.add(
                np.square(np.subtract(grid[:, 0], border_origin[0])),
                np.square(np.subtract(grid[:, 1], border_origin[1])),
            )
        )

        for pixel_index in range(grid.shape[0]):

            if grid_radii[pixel_index] > border_min_radii:

                closest_pixel_index = np.argmin(
                    np.square(grid[pixel_index, 0] - border_grid[:, 0])
                    + np.square(grid[pixel_index, 1] - border_grid[:, 1])
                )

                move_factor = (
                    border_grid_radii[closest_pixel_index] / grid_radii[pixel_index]
                )

                if move_factor < 1.0:
                    grid[pixel_index, :] = (
                        move_factor * (grid[pixel_index, :] - border_origin[:])
                        + border_origin[:]
                    )

        return grid

    @property
    @array_util.Memoizer()
    def sub_mask_1d_index_to_mask_1d_index(self):
        """The mapping_util between every sub-pixel and its host pixel.

        For example:

        - sub_to_pixel[8] = 2 -  The ninth sub-pixel is within the 3rd pixel.
        - sub_to_pixel[20] = 4 -  The twenty first sub-pixel is within the 5th pixel.
        """
        return self.mask.sub_mask_1d_index_to_mask_1d_index_from_sub_grid_size(
            sub_grid_size=self.sub_grid_size
        )

    @property
    @array_util.Memoizer()
    def in_radians(self):
        return (self * np.pi) / 648000.0

    @property
    def masked_shape_arcsec(self):
        return (
            np.amax(self[:, 0]) - np.amin(self[:, 0]),
            np.amax(self[:, 1]) - np.amin(self[:, 1]),
        )

    @property
    def yticks(self):
        """Compute the yticks labels of this grid, used for plotting the y-axis ticks when visualizing a grid"""
        return np.linspace(np.min(self[:, 0]), np.max(self[:, 0]), 4)

    @property
    def xticks(self):
        """Compute the xticks labels of this grid, used for plotting the x-axis ticks when visualizing a grid"""
        return np.linspace(np.min(self[:, 1]), np.max(self[:, 1]), 4)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(Grid, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(Grid, self).__setstate__(state[0:-1])


class BinnedGrid(Grid):
    def __init__(
        self,
        arr,
        mask,
        bin_up_factor,
        binned_mask_1d_index_to_mask_1d_indexes,
        binned_mask_1d_index_to_mask_1d_sizes,
        total_unbinned_pixels,
    ):
        # noinspection PyArgumentList
        super(BinnedGrid, self).__init__()
        self.mask = mask
        self.bin_up_factor = bin_up_factor
        self.binned_mask_1d_index_to_mask_1d_indexes = (
            binned_mask_1d_index_to_mask_1d_indexes
        )
        self.binned_mask_1d_index_to_mask_1d_sizes = (
            binned_mask_1d_index_to_mask_1d_sizes
        )
        self.total_unbinned_pixels = total_unbinned_pixels

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        if isinstance(obj, BinnedGrid):
            self.mask = obj.mask
            self.bin_up_factor = obj.bin_up_factor
            self.binned_mask_1d_index_to_mask_1d_indexes = (
                obj.binned_mask_1d_index_to_mask_1d_indexes
            )
            self.binned_mask_1d_index_to_mask_1d_sizes = (
                obj.binned_mask_1d_index_to_mask_1d_sizes
            )
            self.total_unbinned_pixels = obj.total_unbinned_pixels

    @classmethod
    def from_mask_and_pixel_scale_binned_grid(cls, mask, pixel_scale_binned_grid):

        if pixel_scale_binned_grid > mask.pixel_scale:

            bin_up_factor = int(pixel_scale_binned_grid / mask.pixel_scale)

        else:

            bin_up_factor = 1

        binned_mask = mask.binned_up_mask_from_mask(bin_up_factor=bin_up_factor)

        binned_grid = Grid.from_mask_and_sub_grid_size(
            mask=binned_mask, sub_grid_size=1
        )
        binned_mask_1d_index_to_mask_1d_indexes, binned_mask_1d_index_to_mask_1d_sizes = binning_util.binned_masked_array_1d_to_masked_array_1d_all_from_mask_2d_and_bin_up_factor(
            mask_2d=mask, bin_up_factor=bin_up_factor
        )

        return BinnedGrid(
            arr=binned_grid,
            mask=binned_mask,
            bin_up_factor=bin_up_factor,
            binned_mask_1d_index_to_mask_1d_indexes=binned_mask_1d_index_to_mask_1d_indexes.astype(
                "int"
            ),
            binned_mask_1d_index_to_mask_1d_sizes=binned_mask_1d_index_to_mask_1d_sizes.astype(
                "int"
            ),
            total_unbinned_pixels=mask.pixels_in_mask,
        )


class PixelizationGrid(np.ndarray):
    def __new__(
        cls, arr, mask_1d_index_to_nearest_pixelization_1d_index, *args, **kwargs
    ):
        """A pixelization-grid of (y,x) coordinates which are used to form the pixel centres of adaptive pixelizations in the \
        *pixelizations* module.

        A *PixGrid* is ordered such pixels begin from the top-row of the mask and go rightwards and then \
        downwards. Therefore, it is a ndarray of shape [total_pix_pixels, 2]. The first element of the ndarray \
        thus corresponds to the pixelization pixel index and second element the y or x arc -econd coordinates. For example:

        - pix_grid[3,0] = the 4th unmasked pixel's y-coordinate.
        - pix_grid[6,1] = the 7th unmasked pixel's x-coordinate.

        Parameters
        -----------
        pix_grid : ndarray
            The grid of (y,x) arc-second coordinates of every image-plane pixelization grid used for adaptive source \
            -plane pixelizations.
        mask_1d_index_to_nearest_pixelization_1d_index : ndarray
            A 1D array that maps every grid pixel to its nearest pixelization-grid pixel.
        """
        obj = arr.view(cls)
        obj.mask_1d_index_to_nearest_pixelization_1d_index = (
            mask_1d_index_to_nearest_pixelization_1d_index
        )
        obj.interpolator = None
        return obj

    @classmethod
    def from_grid_and_unmasked_2d_grid_shape(cls, unmasked_sparse_shape, grid):

        sparse_grid = SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=unmasked_sparse_shape, grid=grid
        )

        return PixelizationGrid(
            arr=sparse_grid.sparse,
            mask_1d_index_to_nearest_pixelization_1d_index=sparse_grid.mask_1d_index_to_sparse_1d_index,
        )

    def __array_finalize__(self, obj):
        if hasattr(obj, "mask_1d_index_to_nearest_pixelization_1d_index"):
            self.mask_1d_index_to_nearest_pixelization_1d_index = (
                obj.mask_1d_index_to_nearest_pixelization_1d_index
            )
        if hasattr(obj, "interpolator"):
            self.interpolator = obj.interpolator


class SparseToGrid(scaled_array.RectangularArrayGeometry):
    def __init__(self, sparse_grid, mask_1d_index_to_sparse_1d_index):
        """A sparse grid of coordinates, where each entry corresponds to the (y,x) coordinates at the centre of a \
        pixel on the sparse grid. To setup the sparse-grid, it is laid over a grid of unmasked pixels, such \
        that all sparse-grid pixels which map inside of an unmasked grid pixel are included on the sparse grid.

        To setup this sparse grid, we thus have two sparse grid:

        - The unmasked sparse-grid, which corresponds to a uniform 2D array of pixels. The edges of this grid \
          correspond to the 4 edges of the mask (e.g. the higher and lowest (y,x) arc-second unmasked pixels) and the \
          grid's shape is speciifed by the unmasked_sparse_grid_shape parameter.

        - The (masked) sparse-grid, which is all pixels on the unmasked sparse-grid above which fall within unmasked \
          grid pixels. These are the pixels which are actually used for other modules in PyAutoLens.

        The origin of the unmasked sparse grid can be changed to allow off-center pairings with sparse-grid pixels, \
        which is necessary when a mask has a centre offset from (0.0", 0.0"). However, the sparse grid itself \
        retains an origin of (0.0", 0.0"), ensuring its arc-second grid uses the same coordinate system as the \
        other grid.

        The sparse grid is used to determine the pixel centers of an adaptive grid pixelization.

        Parameters
        ----------
        unmasked_sparse_shape : (int, int)
            The shape of the unmasked sparse-grid whose centres form the sparse-grid.
        pixel_scales : (float, float)
            The pixel-to-arcsecond scale of a pixel in the y and x directions.
        grid : Grid
            The grid used to determine which pixels are in the sparse grid.
        origin : (float, float)
            The centre of the unmasked sparse grid, which matches the centre of the mask.
        """
        self.sparse = sparse_grid
        self.mask_1d_index_to_sparse_1d_index = mask_1d_index_to_sparse_1d_index

    @classmethod
    def from_grid_and_unmasked_2d_grid_shape(cls, grid, unmasked_sparse_shape):
        """Calculate the image-plane pixelization from a grid of coordinates (and its mask).

        See *grid_stacks.SparseToGrid* for details on how this grid is calculated.

        Parameters
        -----------
        grid : grids.Grid
            The grid of (y,x) arc-second coordinates at the centre of every image value (e.g. image-pixels).
        """

        pixel_scale = grid.mask.pixel_scale

        pixel_scales = (
            (grid.masked_shape_arcsec[0] + pixel_scale) / (unmasked_sparse_shape[0]),
            (grid.masked_shape_arcsec[1] + pixel_scale) / (unmasked_sparse_shape[1]),
        )

        origin = grid.mask.centre

        unmasked_sparse_grid = grid_util.grid_1d_from_shape_pixel_scales_sub_grid_size_and_origin(
            shape=unmasked_sparse_shape,
            pixel_scales=pixel_scales,
            sub_grid_size=1,
            origin=origin,
        )

        unmasked_sparse_grid_pixel_centres = grid.mask.grid_arcsec_to_grid_pixel_centres(
            grid_arcsec=unmasked_sparse_grid
        )

        total_sparse_pixels = mask_util.total_sparse_pixels_from_mask(
            mask=grid.mask,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        unmasked_sparse_to_sparse = sparse_mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
            mask=grid.mask,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            total_sparse_pixels=total_sparse_pixels,
        ).astype(
            "int"
        )

        sparse_to_unmasked_sparse = sparse_mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_sparse_pixels,
            mask=grid.mask,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        ).astype(
            "int"
        )

        regular_to_unmasked_sparse = grid_util.grid_arcsec_1d_to_grid_pixel_indexes_1d(
            grid_arcsec_1d=grid.unlensed_unsubbed_1d,
            shape=unmasked_sparse_shape,
            pixel_scales=pixel_scales,
            origin=origin,
        ).astype("int")

        mask_1d_index_to_sparse_1d_index = sparse_mapping_util.mask_1d_index_to_sparse_1d_index_from_sparse_mappings(
            regular_to_unmasked_sparse=regular_to_unmasked_sparse,
            unmasked_sparse_to_sparse=unmasked_sparse_to_sparse,
        ).astype(
            "int"
        )

        sparse_grid = sparse_mapping_util.sparse_grid_from_unmasked_sparse_grid(
            unmasked_sparse_grid=unmasked_sparse_grid,
            sparse_to_unmasked_sparse=sparse_to_unmasked_sparse,
        )

        return SparseToGrid(
            sparse_grid=sparse_grid,
            mask_1d_index_to_sparse_1d_index=mask_1d_index_to_sparse_1d_index,
        )

    @classmethod
    def from_total_pixels_binned_grid_and_weight_map(
        cls,
        total_pixels,
        binned_grid,
        binned_weight_map,
        n_iter=1,
        max_iter=5,
        seed=None,
    ):
        """Calculate the image-plane pixelization from a grid of coordinates (and its mask).

        See *grid_stacks.SparseToGrid* for details on how this grid is calculated.

        Parameters
        -----------
        binned_grid : grids.Grid
            The grid of (y,x) arc-second coordinates at the centre of every image value (e.g. image-pixels).
        """

        if total_pixels > binned_grid.shape[0]:
            raise exc.PixelizationException

        kmeans = KMeans(
            n_clusters=total_pixels, random_state=seed, n_init=n_iter, max_iter=max_iter
        )

        kmeans = kmeans.fit(X=binned_grid, sample_weight=binned_weight_map)

        mask_1d_index_to_sparse_1d_index = sparse_mapping_util.mask_1d_index_to_sparse_1d_index_from_binned_grid(
            sparse_labels=kmeans.labels_,
            binned_mask_1d_index_to_mask_1d_indexes=binned_grid.binned_mask_1d_index_to_mask_1d_indexes,
            binned_mask_1d_index_to_mask_1d_sizes=binned_grid.binned_mask_1d_index_to_mask_1d_sizes,
            total_unbinned_pixels=binned_grid.total_unbinned_pixels,
        )

        return SparseToGrid(
            sparse_grid=kmeans.cluster_centers_,
            mask_1d_index_to_sparse_1d_index=mask_1d_index_to_sparse_1d_index.astype(
                "int"
            ),
        )

    @property
    def total_sparse_pixels(self):
        return len(self.sparse)


class Interpolator(object):
    def __init__(self, grid, interp_grid, pixel_scale_interpolation_grid):
        self.grid = grid
        self.interp_grid = interp_grid
        self.pixel_scale_interpolation_grid = pixel_scale_interpolation_grid
        self.vtx, self.wts = self.interp_weights

    @property
    def interp_weights(self):
        tri = qhull.Delaunay(self.interp_grid)
        simplex = tri.find_simplex(self.grid)
        # noinspection PyUnresolvedReferences
        vertices = np.take(tri.simplices, simplex, axis=0)
        temp = np.take(tri.transform, simplex, axis=0)
        delta = self.grid - temp[:, 2]
        bary = np.einsum("njk,nk->nj", temp[:, :2, :], delta)
        return vertices, np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))

    @classmethod
    def from_mask_grid_and_pixel_scale_interpolation_grids(
        cls, mask, grid, pixel_scale_interpolation_grid
    ):

        rescale_factor = mask.pixel_scale / pixel_scale_interpolation_grid

        rescaled_mask = mask_util.rescaled_mask_2d_from_mask_2d_and_rescale_factor(
            mask_2d=mask, rescale_factor=rescale_factor
        )

        interp_mask = mask_util.edge_buffed_mask_from_mask(mask=rescaled_mask).astype(
            "bool"
        )

        interp_grid = grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
            mask=interp_mask,
            pixel_scales=(
                pixel_scale_interpolation_grid,
                pixel_scale_interpolation_grid,
            ),
            sub_grid_size=1,
            origin=mask.origin,
        )

        return Interpolator(
            grid=grid,
            interp_grid=interp_grid,
            pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        )

    def interpolated_values_from_values(self, values):
        return np.einsum("nj,nj->n", np.take(values, self.vtx), self.wts)


def grid_interpolate(func):
    """
    Decorate a profile method that accepts a coordinate grid and returns a instrument grid.

    If an interpolator attribute is associated with the input grid then that interpolator is used to down sample the
    coordinate grid prior to calling the function and up sample the result of the function.

    If no interpolator attribute is associated with the input grid then the function is called as hyper.

    Parameters
    ----------
    func
        Some method that accepts a grid

    Returns
    -------
    decorated_function
        The function with optional interpolation
    """

    @wraps(func)
    def wrapper(profile, grid, grid_radial_minimum=None, *args, **kwargs):
        if hasattr(grid, "interpolator"):
            interpolator = grid.interpolator
            if grid.interpolator is not None:
                values = func(
                    profile,
                    interpolator.interp_grid,
                    grid_radial_minimum,
                    *args,
                    **kwargs
                )
                if values.ndim == 1:
                    return interpolator.interpolated_values_from_values(values=values)
                elif values.ndim == 2:
                    y_values = interpolator.interpolated_values_from_values(
                        values=values[:, 0]
                    )
                    x_values = interpolator.interpolated_values_from_values(
                        values=values[:, 1]
                    )
                    return np.asarray([y_values, x_values]).T
        return func(profile, grid, grid_radial_minimum, *args, **kwargs)

    return wrapper



import autofit as af
import autolens as al
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging

# We've learnt nearly all the tools we need to model strong lenses, so I'm now going to quickly cover how you should
# choose your mask. I'll also show you another neat trick to improve the speed and accuracy of your non-linear search.

# You need to change the path below to the chapter 2 directory.
chapter_path = "/path/to/user/autolens_workspace/howtolens/chapter_2_lens_modeling/"
chapter_path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/howtolens/chapter_2_lens_modeling/"

af.conf.instance = af.conf.Config(
    config_path=chapter_path + "/configs/t8_masking_and_positions",
    output_path=chapter_path + "/output",
)

# Lets simulate the simple image we've used throughout this chapter.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=0.1)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(130, 130), pixel_scale=0.1, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.SphericalExponential(
            centre=(0.0, 0.0), intensity=0.2, effective_radius=0.2
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.1,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Simulate the CCD data.
ccd_data = simulate()
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# When it comes to determining an appropriate mask for this image, the best approach is to set up a mask using the mask
# module and pass it to a ccd plotter. You can then check visually if the mask is an appropriate size or not.
# Below, we choose an inner radius that cuts into our lensed source galaxy - clearly this isn't a good mask.
mask = al.Mask.circular_annular(
    shape=ccd_data.shape,
    pixel_scale=ccd_data.pixel_scale,
    inner_radius_arcsec=1.4,
    outer_radius_arcsec=2.4,
)

al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data, mask=mask, extract_array_from_mask=True, zoom_around_mask=True
)

# So, lets decrease the inner radius to correct for this.
mask = al.Mask.circular_annular(
    shape=ccd_data.shape,
    pixel_scale=ccd_data.pixel_scale,
    inner_radius_arcsec=0.6,
    outer_radius_arcsec=2.4,
)

al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data, mask=mask, extract_array_from_mask=True, zoom_around_mask=True
)

# When we run the phase, we don't pass it the mask as an array. Instead, we pass it the mask as a function. The reason
# for this will become clear in the next chapter, but for now I would say you just accept this syntax.
def mask_function():
    return al.Mask.circular_annular(
        shape=ccd_data.shape,
        pixel_scale=ccd_data.pixel_scale,
        inner_radius_arcsec=0.6,
        outer_radius_arcsec=2.4,
    )


phase_with_custom_mask = phase_imaging.PhaseImaging(
    phase_name="phase_t8_with_custom_mask",
    galaxies=dict(
        lens=gm.GalaxyModel(redshift=0.5), source=gm.GalaxyModel(redshift=1.0)
    ),
    mask_function=mask_function,  # <- We input the mask function here
    optimizer_class=af.MultiNest,
)

# So, our mask encompasses the lensed source galaxy. However, is this really the right sized mask? Do we want
# a bigger mask? a smaller mask?

# When it comes to masking, we are essentially balancing run-speed and accuracy. If speed wasn't a consideration,
# bigger masks would *always* be better, for two reasons:

# 1) The lensed source galaxy may have very faint emission that when you look at the plot above you don't notice.
#    Overly aggressive masking risks you masking out some of that light - data which would better constrain your
#    lens model!

# 2) When you fit an image with a model image the fit is performed only within the masked region. Outside of the
#    masked region it is possible that the model image produces some source-galaxy light in a region of the image
#    where it isn't actually observed. If this region is masked, the poor fit in this region won't reduce the model's
#    likelihood.

# As you use PyAutoLens more you will get a feel for how fast an analysis will run given a certain image resolution,
# lens model complexity, non-linear search priors / settings, etc. As you develop this intuition, I would recommend you
# always aim to use masks as big as possible which still give a reasonable run-speed. Aggresive masking
# will get your code running fast - but it could lead you to infer an incorrect lens model!

# If you are fitting the foreground lens galaxy's light you pretty much have no choice but to use a large circular
# mask anyway, as you'll need to capture the lens's extended emission. Chances are this will encompass the entire
# source galaxy.


# We can also manually specify a set of image-pixels which correspond to the multiple images of the source-galaxy(s).
# During the analysis, PyAutoLens will first check that these pixels trace within a specified arc-second threshold of
# one another (which is controlled by the 'position_threshold' parameter input into a phase). This
# provides two benefits:

# 1) The analysis runs faster as the non-linear search avoids searching regions of parameter space where the
#    mass-model is clearly not accurate.

# 2) By removing these solutions, a global-maximum solution may be reached instead of a local-maxima. This is because
#    removing the incorrect mass models makes the non-linear parameter space less complex.

# We can easily check the image-positions are accurate by plotting them using our ccd plotter (they are the magenta
# dots on the image).
al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data, positions=[[[1.6, 0.0], [0.0, 1.6], [-1.6, 0.0], [0.0, -1.6]]]
)

# We can then tell our phase to use these positions in the analysis.
phase_with_positions = phase_imaging.PhaseImaging(
    phase_name="phase_t8_with_positions",
    galaxies=dict(
        lens=gm.GalaxyModel(redshift=0.5, mass=al.mass_profiles.SphericalIsothermal),
        source=gm.GalaxyModel(
            redshift=1.0, light=al.light_profiles.SphericalExponential
        ),
    ),
    positions_threshold=0.5,  # <- We input a positions threshold here, to signify how far pixels must trace within one another.
    optimizer_class=af.MultiNest,
)

# The positions are passed to the phase when we run it,as shown below.
print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t7_multinest_black_magic"
    " folder for live output of the results, images and lens model."
    " This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)

phase_with_positions.run(
    data=ccd_data, positions=[[[1.6, 0.0], [0.0, 1.6], [-1.6, 0.0], [0.0, -1.6]]]
)

print("MultiNest has finished run - you may now continue the notebook.")
# You may observe multiple source-galaxies each with their own set of multiple-images. If you have a means by
# which to pair different positions to the same source galaxies (for example, spectroscopic data) you can set up
# multiple sets of positions which each have to trace to within the position threshold of one another for the lens
# model to be accepted.


def simulate_two_galaxies():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=0.1)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(130, 130), pixel_scale=0.1, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6
        ),
    )

    source_galaxy_0 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.SphericalExponential(
            centre=(1.0, 0.0), intensity=0.2, effective_radius=0.2
        ),
    )

    source_galaxy_1 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.SphericalExponential(
            centre=(-1.0, 0.0), intensity=0.2, effective_radius=0.2
        ),
    )

    tracer = al.Tracer.from_galaxies(
        galaxies=[lens_galaxy, source_galaxy_0, source_galaxy_1]
    )

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.1,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Simulate the new data.

ccd_data = simulate_two_galaxies()

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# To specify the positions we break the positions list into two cells. They will be plotted in different colours to
# represent the fact they trace from different source galaxies.
al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data,
    positions=[[[2.65, 0.0], [-0.55, 0.0]], [[-2.65, 0.0], [0.55, 0.0]]],
)

# Again, we tell our phase to use the positions and pass this list of pixels to our phase when we run it.
phase_with_x2_positions = phase_imaging.PhaseImaging(
    phase_name="phase_t8_with_x2_positions",
    galaxies=dict(
        lens=gm.GalaxyModel(redshift=0.5, mass=al.mass_profiles.SphericalIsothermal),
        source_0=gm.GalaxyModel(
            redshift=1.0, light=al.light_profiles.SphericalExponential
        ),
        source_1=gm.GalaxyModel(
            redshift=1.0, light=al.light_profiles.SphericalExponential
        ),
    ),
    positions_threshold=0.5,  # <- We input a positions threshold here, to signify how far pixels must trace within one another.
    optimizer_class=af.MultiNest,
)

phase_with_x2_positions.run(
    data=ccd_data, positions=[[[2.65, 0.0], [-0.55, 0.0]], [[-2.65, 0.0], [0.55, 0.0]]]
)

# And that completes our final tutorial in this chapter! At this point, I recommend that you checkout the
# 'autolens_workspace/tools/mask_maker.py' and 'autolens_workspace/tools/positions_maker.py'
# scripts. These tools allow you create custom masks and positions for a specific strong lens and output them so
# they can be loaded before an analysis.

# When we cover pipelines next, you'll see that pipelines allow us to use a custom mask and set of positions for each lens
# we model. So, although we have to draw the masks and positions for each lens in a sample, once we've done that
# we can fit all lenses with one standardized pipeline!

#  There are two things you should bare in mind in terms of masking and positions:
#
# 1) Customizing the mask and positions for the analysis of one strong lens gets the analysis running fast and can
#    provide accurate non-linear sampling. However, for a large sample of lenses, customizing the mask and positions
#    will begin to take a lot of time. If you're willing to put that time and effort in, great, but these solutions
#    *do not* scale-up to large samples of lenses.

# 2) A word of warning - be *extremely* careful when using positions, especially if it is unclear if the lensed source
#    galaxy has one or multiple source's of light. If your position threshold is small and the positions you give the
#    analysis correspond to different parts of the source, you may remove the *correct lens model*. In my experience,
#    as long as you keep the threshold above ~0.5" you'll be fine.

# And with that, we've completed the chapter.



from astropy import cosmology
from astropy import constants
from astropy import units
import autolens as al

import numpy as np

from astropy import cosmology

cosmo = cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)

z_nfw = 0.6
z_source = 2.5

nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.119, scale_radius=0.052)

concentration = nfw.concentration_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

mass_at_200 = nfw.mass_at_200_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

print("Concentration = ", concentration)
print("Mass at 200 = {:.4e}".format(mass_at_200))

nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.08, scale_radius=0.3)

concentration = nfw.concentration_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

mass_at_200 = nfw.mass_at_200_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

print("Concentration = ", concentration)
print("Mass at 200 = {:.4e}".format(mass_at_200))

nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.01, scale_radius=0.07)

concentration = nfw.concentration_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

mass_at_200 = nfw.mass_at_200_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

print("Concentration = ", concentration)
print("Mass at 200 = {:.4e}".format(mass_at_200))


nfw = al.mass_profiles.SphericalTruncatedNFWChallenge(kappa_s=0.002, scale_radius=0.02)

concentration = nfw.concentration_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

mass_at_200 = nfw.mass_at_200_for_units(
    redshift_profile=z_nfw,
    redshift_source=z_source,
    unit_length="kpc",
    unit_mass="solMass",
    redshift_of_cosmic_average_density="profile",
    cosmology=cosmo,
)

print("Concentration = ", concentration)
print("Mass at 200 = {:.4e}".format(mass_at_200))



import copy

import numpy as np
from typing import cast

import autofit as af
from autolens.lens import lens_data as ld, lens_fit
from autolens.model.galaxy import galaxy as g
from autolens.model.hyper import hyper_data as hd
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline.plotters import hyper_plotters
from .hyper_phase import HyperPhase


class HyperGalaxyPhase(HyperPhase):
    def __init__(self, phase):

        super().__init__(phase=phase, hyper_name="hyper_galaxy")
        self.include_sky_background = False
        self.include_noise_background = False

    class Analysis(af.Analysis):
        def __init__(
            self, lens_data, hyper_model_image_1d, hyper_galaxy_image_1d_path_dict
        ):
            """
            An analysis to fit the noise for a single galaxy image.
            Parameters
            ----------
            lens_data: LensData
                lens data, including an image and noise
            hyper_model_image_1d: ndarray
                An image produce of the overall system by a model
            hyper_galaxy_image_1d_path_dict: ndarray
                The contribution of one galaxy to the model image
            """

            self.lens_data = lens_data

            self.hyper_model_image_1d = hyper_model_image_1d
            self.hyper_galaxy_image_1d_path_dict = hyper_galaxy_image_1d_path_dict

            self.plot_hyper_galaxy_subplot = af.conf.instance.visualize.get(
                "plots", "plot_hyper_galaxy_subplot", bool
            )

        def visualize(self, instance, image_path, during_analysis):

            if self.plot_hyper_galaxy_subplot:

                subplot_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                    path=image_path, folder_names=["subplots"]
                )

                hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

                hyper_background_noise = self.hyper_background_noise_for_instance(
                    instance=instance
                )

                hyper_model_image_2d = self.lens_data.grid.scaled_array_2d_from_array_1d(
                    array_1d=self.hyper_model_image_1d
                )

                hyper_galaxy_image_2d = self.lens_data.grid.scaled_array_2d_from_array_1d(
                    array_1d=self.hyper_galaxy_image_1d_path_dict
                )

                contribution_map_2d = instance.hyper_galaxy.contribution_map_from_hyper_images(
                    hyper_model_image=hyper_model_image_2d,
                    hyper_galaxy_image=hyper_galaxy_image_2d,
                )

                fit_normal = lens_fit.LensDataFit(
                    lens_data=self.lens_data,
                    image_1d=self.lens_data.image_1d,
                    noise_map_1d=self.lens_data.noise_map_1d,
                    mask_1d=self.lens_data.mask_1d,
                    model_image_1d=self.hyper_model_image_1d,
                )

                fit = self.fit_for_hyper_galaxy(
                    hyper_galaxy=instance.hyper_galaxy,
                    hyper_image_sky=hyper_image_sky,
                    hyper_background_noise=hyper_background_noise,
                )

                hyper_plotters.plot_hyper_galaxy_subplot(
                    hyper_galaxy_image=hyper_galaxy_image_2d,
                    contribution_map=contribution_map_2d,
                    noise_map=self.lens_data.noise_map(return_in_2d=True),
                    hyper_noise_map=fit.noise_map(return_in_2d=True),
                    chi_squared_map=fit_normal.chi_squared_map(return_in_2d=True),
                    hyper_chi_squared_map=fit.chi_squared_map(return_in_2d=True),
                    output_path=subplot_path,
                    output_format="png",
                )

        def fit(self, instance):
            """
            Fit the model image to the real image by scaling the hyper_galaxies noise.
            Parameters
            ----------
            instance: ModelInstance
                A model instance with a hyper_galaxies galaxy property
            Returns
            -------
            fit: float
            """

            hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

            hyper_background_noise = self.hyper_background_noise_for_instance(
                instance=instance
            )

            fit = self.fit_for_hyper_galaxy(
                hyper_galaxy=instance.hyper_galaxy,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            return fit.figure_of_merit

        @staticmethod
        def hyper_image_sky_for_instance(instance):
            if hasattr(instance, "hyper_image_sky"):
                return instance.hyper_image_sky

        @staticmethod
        def hyper_background_noise_for_instance(instance):
            if hasattr(instance, "hyper_background_noise"):
                return instance.hyper_background_noise

        def fit_for_hyper_galaxy(
            self, hyper_galaxy, hyper_image_sky, hyper_background_noise
        ):

            image_1d = lens_fit.image_1d_from_lens_data_and_hyper_image_sky(
                lens_data=self.lens_data, hyper_image_sky=hyper_image_sky
            )

            if hyper_background_noise is not None:
                noise_map_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
                    noise_map=self.lens_data.noise_map_1d
                )
            else:
                noise_map_1d = self.lens_data.noise_map_1d

            hyper_noise_map_1d = hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                hyper_model_image=self.hyper_model_image_1d,
                hyper_galaxy_image=self.hyper_galaxy_image_1d_path_dict,
                noise_map=self.lens_data.noise_map_1d,
            )

            noise_map_1d = noise_map_1d + hyper_noise_map_1d
            if self.lens_data.hyper_noise_map_max is not None:
                noise_map_1d[
                    noise_map_1d > self.lens_data.hyper_noise_map_max
                ] = self.lens_data.hyper_noise_map_max

            return lens_fit.LensDataFit(
                lens_data=self.lens_data,
                image_1d=image_1d,
                noise_map_1d=noise_map_1d,
                mask_1d=self.lens_data.mask_1d,
                model_image_1d=self.hyper_model_image_1d,
            )

        @classmethod
        def describe(cls, instance):
            return "Running hyper_galaxies galaxy fit for HyperGalaxy:\n{}".format(
                instance.hyper_galaxy
            )

    def run_hyper(self, data, results=None):
        """
        Run a fit for each galaxy from the previous phase.
        Parameters
        ----------
        data: LensData
        results: ResultsCollection
            Results from all previous phases
        Returns
        -------
        results: HyperGalaxyResults
            A collection of results, with one item per a galaxy
        """
        phase = self.make_hyper_phase()

        lens_data = ld.LensData(
            ccd_data=data,
            mask=results.last.mask_2d,
            sub_grid_size=cast(phase_imaging.PhaseImaging, phase).sub_grid_size,
            trimmed_psf_shape=cast(phase_imaging.PhaseImaging, phase).psf_shape,
            positions=results.last.positions,
            positions_threshold=cast(
                phase_imaging.PhaseImaging, phase
            ).positions_threshold,
            pixel_scale_interpolation_grid=cast(
                phase_imaging.PhaseImaging, phase
            ).pixel_scale_interpolation_grid,
            pixel_scale_binned_grid=cast(
                phase_imaging.PhaseImaging, phase
            ).pixel_scale_binned_cluster_grid,
            inversion_pixel_limit=cast(
                phase_imaging.PhaseImaging, phase
            ).inversion_pixel_limit,
            inversion_uses_border=cast(
                phase_imaging.PhaseImaging, phase
            ).inversion_uses_border,
            hyper_noise_map_max=cast(
                phase_imaging.PhaseImaging, phase
            ).hyper_noise_map_max,
            preload_pixelization_grids_of_planes=None,
        )

        model_image_1d = results.last.hyper_model_image_1d
        hyper_galaxy_image_1d_path_dict = results.last.hyper_galaxy_image_1d_path_dict

        hyper_result = copy.deepcopy(results.last)
        hyper_result.variable = hyper_result.variable.copy_with_fixed_priors(
            hyper_result.constant
        )

        hyper_result.analysis.hyper_model_image_1d = model_image_1d
        hyper_result.analysis.hyper_galaxy_image_1d_path_dict = (
            hyper_galaxy_image_1d_path_dict
        )

        for path, galaxy in results.last.path_galaxy_tuples:

            optimizer = phase.optimizer.copy_with_name_extension(extension=path[-1])

            optimizer.phase_tag = ""

            # TODO : This is a HACK :O

            optimizer.variable.galaxies = []

            optimizer.const_efficiency_mode = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_const_efficiency_mode", bool
            )
            optimizer.sampling_efficiency = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_sampling_efficiency", float
            )
            optimizer.n_live_points = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_n_live_points", int
            )
            optimizer.multimodal = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_multimodal", bool
            )

            optimizer.variable.hyper_galaxy = g.HyperGalaxy

            if self.include_sky_background:
                optimizer.variable.hyper_image_sky = hd.HyperImageSky

            if self.include_noise_background:
                optimizer.variable.hyper_background_noise = hd.HyperBackgroundNoise

            # If array is all zeros, galaxy did not have image in previous phase and
            # should be ignored
            if not np.all(hyper_galaxy_image_1d_path_dict[path] == 0):

                analysis = self.Analysis(
                    lens_data=lens_data,
                    hyper_model_image_1d=model_image_1d,
                    hyper_galaxy_image_1d_path_dict=hyper_galaxy_image_1d_path_dict[
                        path
                    ],
                )

                result = optimizer.fit(analysis)

                def transfer_field(name):
                    if hasattr(result.constant, name):
                        setattr(
                            hyper_result.constant.object_for_path(path),
                            name,
                            getattr(result.constant, name),
                        )
                        setattr(
                            hyper_result.variable.object_for_path(path),
                            name,
                            getattr(result.variable, name),
                        )

                transfer_field("hyper_galaxy")

                hyper_result.constant.hyper_image_sky = getattr(
                    result.constant, "hyper_image_sky"
                )
                hyper_result.variable.hyper_image_sky = getattr(
                    result.variable, "hyper_image_sky"
                )

                hyper_result.constant.hyper_background_noise = getattr(
                    result.constant, "hyper_background_noise"
                )
                hyper_result.variable.hyper_background_noise = getattr(
                    result.variable, "hyper_background_noise"
                )

        return hyper_result


class HyperGalaxyBackgroundSkyPhase(HyperGalaxyPhase):
    def __init__(self, phase):
        super().__init__(phase=phase)
        self.include_sky_background = True
        self.include_noise_background = False


class HyperGalaxyBackgroundNoisePhase(HyperGalaxyPhase):
    def __init__(self, phase):
        super().__init__(phase=phase)
        self.include_sky_background = False
        self.include_noise_background = True


class HyperGalaxyBackgroundBothPhase(HyperGalaxyPhase):
    def __init__(self, phase):
        super().__init__(phase=phase)
        self.include_sky_background = True
        self.include_noise_background = True


class HyperGalaxyAllPhase(HyperPhase):
    def __init__(
        self, phase, include_sky_background=False, include_noise_background=False
    ):
        super().__init__(phase=phase, hyper_name="hyper_galaxy")
        self.include_sky_background = include_sky_background
        self.include_noise_background = include_noise_background

    def run_hyper(self, data, results=None):
        """
        Run a fit for each galaxy from the previous phase.
        Parameters
        ----------
        data: LensData
        results: ResultsCollection
            Results from all previous phases
        Returns
        -------
        results: HyperGalaxyResults
            A collection of results, with one item per a galaxy
        """
        phase = self.make_hyper_phase()

        lens_data = ld.LensData(
            ccd_data=data,
            mask=results.last.mask_2d,
            sub_grid_size=cast(phase_imaging.PhaseImaging, phase).sub_grid_size,
            trimmed_psf_shape=cast(phase_imaging.PhaseImaging, phase).psf_shape,
            positions=results.last.positions,
            positions_threshold=cast(
                phase_imaging.PhaseImaging, phase
            ).positions_threshold,
            pixel_scale_interpolation_grid=cast(
                phase_imaging.PhaseImaging, phase
            ).pixel_scale_interpolation_grid,
            pixel_scale_binned_grid=cast(
                phase_imaging.PhaseImaging, phase
            ).pixel_scale_binned_cluster_grid,
            inversion_pixel_limit=cast(
                phase_imaging.PhaseImaging, phase
            ).inversion_pixel_limit,
            inversion_uses_border=cast(
                phase_imaging.PhaseImaging, phase
            ).inversion_uses_border,
            hyper_noise_map_max=cast(
                phase_imaging.PhaseImaging, phase
            ).hyper_noise_map_max,
            preload_pixelization_grids_of_planes=None,
        )

        model_image_1d = results.last.hyper_model_image_1d
        hyper_galaxy_image_1d_path_dict = results.last.hyper_galaxy_image_1d_path_dict

        hyper_result = copy.deepcopy(results.last)
        hyper_result.variable = hyper_result.variable.copy_with_fixed_priors(
            hyper_result.constant
        )

        hyper_result.analysis.hyper_model_image_1d = model_image_1d
        hyper_result.analysis.hyper_galaxy_image_1d_path_dict = (
            hyper_galaxy_image_1d_path_dict
        )

        for path, galaxy in results.last.path_galaxy_tuples:

            optimizer = phase.optimizer.copy_with_name_extension(extension=path[-1])

            optimizer.phase_tag = ""

            # TODO : This is a HACK :O

            optimizer.variable.galaxies = []
            optimizer.variable.galaxies = []
            optimizer.variable.galaxies = []

            optimizer.const_efficiency_mode = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_const_efficiency_mode", bool
            )
            optimizer.sampling_efficiency = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_sampling_efficiency", float
            )
            optimizer.n_live_points = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_n_live_points", int
            )

            optimizer.variable.hyper_galaxy = g.HyperGalaxy

            if self.include_sky_background:
                optimizer.variable.hyper_image_sky = hd.HyperImageSky

            if self.include_noise_background:
                optimizer.variable.hyper_background_noise = hd.HyperBackgroundNoise

            # If array is all zeros, galaxy did not have image in previous phase and
            # should be ignored
            if not np.all(hyper_galaxy_image_1d_path_dict[path] == 0):

                analysis = self.Analysis(
                    lens_data=lens_data,
                    model_image_1d=model_image_1d,
                    galaxy_image_1d=hyper_galaxy_image_1d_path_dict[path],
                )

                result = optimizer.fit(analysis)

                def transfer_field(name):
                    if hasattr(result.constant, name):
                        setattr(
                            hyper_result.constant.object_for_path(path),
                            name,
                            getattr(result.constant, name),
                        )
                        setattr(
                            hyper_result.variable.object_for_path(path),
                            name,
                            getattr(result.variable, name),
                        )

                transfer_field("hyper_galaxy")
                transfer_field("hyper_image_sky")
                transfer_field("hyper_background_noise")

        return hyper_result



import autofit as af
import autolens as al

import os

# This tool allows one to mask a bespoke noise-map for a given image of a strong lens, which can then be loaded before a
# pipeline is run and passed to that pipeline so as to become the default masked used by a phase (if a mask
# function is not passed to that phase).

# This noise-map is primarily used for increasing the variances of pixels that have non-modeled components in an image,
# for example intervening line-of-sight galaxies that are near the lens, but not directly interfering with the
# analysis of the lens and source galaxies.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# The 'data name' is the name of the data folder and 'data_name' the folder the mask is stored in, e.g,
# the mask will be output as '/workspace/data/data_type/data_name/mask.fits'.
data_type = "example"
data_name = "lens_sie__source_sersic__intervening_objects"

# Create the path where the noise-map will be output, which in this case is
# '/workspace/data/example/lens_light_and_x1_source_intervening_objects/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

# If you use this tool for your own instrument, you *must* double check this pixel scale is correct!
pixel_scale = 0.1

# First, load the CCD imaging data, so that the location of galaxies is clear when scaling the noise-map.
image = al.load_image(
    image_path=data_path + "image.fits", image_hdu=0, pixel_scale=pixel_scale
)

# Next, load the CCD imaging noise-map, which we will use the scale the noise-map.
noise_map = al.load_noise_map(
    noise_map_path=data_path + "noise_map.fits",
    noise_map_hdu=0,
    pixel_scale=pixel_scale,
)

# Now lets plot the image and mask, so we can check that the mask includes the regions of the image we want.
# data_plotters.plot_signal_to_noise_map(signal_to_noise_map=image / noise_map)

# Here, we manually increase the noise values to extremely large values, such that the analysis essentially omits them.
noise_map[25:55, 77:96] = 1.0e8
noise_map[55:85, 3:27] = 1.0e8

# The signal to noise map is the best way to determine if these regions are appropriately masked out.
al.data_plotters.plot_signal_to_noise_map(signal_to_noise_map=image / noise_map)

# Now we're happy with the mask, lets output it to the data folder of the lens, so that we can load it from a .fits
# file in our pipelines!
al.array_util.numpy_array_2d_to_fits(
    array_2d=noise_map, file_path=data_path + "noise_map_scaled.fits", overwrite=True
)



import numpy as np
from autolens import decorator_util


@decorator_util.jit()
def mapping_matrix_from_sub_mask_1d_index_to_pixelization_1d_index(
    sub_mask_1d_index_to_pixelization_1d_index,
    pixels,
    total_mask_pixels,
    sub_mask_1d_index_to_mask_1d_index,
    sub_grid_fraction,
):
    """Computes the mapping_util matrix, by iterating over the known mappings between the sub-grid and pixelization.

    Parameters
    -----------
    sub_mask_1d_index_to_pixelization_1d_index : ndarray
        The mappings between the observed grid's sub-pixels and pixelization's pixels.
    pixels : int
        The number of pixels in the pixelization.
    total_mask_pixels : int
        The number of datas pixels in the observed datas and thus on the grid.
    sub_mask_1d_index_to_mask_1d_index : ndarray
        The mappings between the observed grid's sub-pixels and observed grid's pixels.
    sub_grid_fraction : float
        The fractional area each sub-pixel takes up in an pixel.
    """

    mapping_matrix = np.zeros((total_mask_pixels, pixels))

    for sub_mask_1d_index in range(sub_mask_1d_index_to_mask_1d_index.shape[0]):
        mapping_matrix[
            sub_mask_1d_index_to_mask_1d_index[sub_mask_1d_index],
            sub_mask_1d_index_to_pixelization_1d_index[sub_mask_1d_index],
        ] += sub_grid_fraction

    return mapping_matrix


@decorator_util.jit()
def voronoi_sub_mask_1d_index_to_pixeliztion_1d_index_from_grids_and_geometry(
    grid,
    mask_1d_index_to_nearest_pixelization_1d_index,
    sub_mask_1d_index_to_mask_1d_index,
    pixel_centres,
    pixel_neighbors,
    pixel_neighbors_size,
):
    """ Compute the mappings between a set of sub-grid pixels and pixelization pixels, using information on \
    how the pixels hosting each sub-pixel map to their closest pixelization pixel on the image-plane pix-grid \
    and the pixelization's pixel centres.

    To determine the complete set of sub-pixel to pixelization pixel mappings, we must pair every sub-pixel to \
    its nearest pixel. Using a full nearest neighbor search to do this is slow, thus the pixel neighbors (derived via \
    the Voronoi grid) are used to localize each nearest neighbor search by using a graph search.

    Parameters
    ----------
    grid : Grid
        The grid of (y,x) arc-second coordinates at the centre of every unmasked pixel, which has been traced to \
        to an irgrid via lens.
    mask_1d_index_to_nearest_pixelization_1d_index : ndarray
        A 1D array that maps every grid pixel to its nearest pix-grid pixel (as determined on the unlensed \
        2D array).
    pixel_centres : (float, float)
        The (y,x) centre of every Voronoi pixel in arc-seconds.
    pixel_neighbors : ndarray
        An array of length (voronoi_pixels) which provides the index of all neighbors of every pixel in \
        the Voronoi grid (entries of -1 correspond to no neighbor).
    pixel_neighbors_size : ndarray
        An array of length (voronoi_pixels) which gives the number of neighbors of every pixel in the \
        Voronoi grid.
     """

    sub_mask_1d_index_to_pixeliztion_1d_index = np.zeros((grid.shape[0]))

    for sub_mask_1d_index in range(grid.shape[0]):

        nearest_pixelization_1d_index = mask_1d_index_to_nearest_pixelization_1d_index[
            sub_mask_1d_index_to_mask_1d_index[sub_mask_1d_index]
        ]

        while True:

            nearest_pixelization_pixel_center = pixel_centres[
                nearest_pixelization_1d_index
            ]

            sub_pixel_to_nearest_pixelization_distance = (
                (grid[sub_mask_1d_index, 0] - nearest_pixelization_pixel_center[0]) ** 2
                + (grid[sub_mask_1d_index, 1] - nearest_pixelization_pixel_center[1])
                ** 2
            )

            closest_separation_from_pixelization_to_neighbor = 1.0e8

            for neighbor_pixelization_1d_index in range(
                pixel_neighbors_size[nearest_pixelization_1d_index]
            ):

                neighbor = pixel_neighbors[
                    nearest_pixelization_1d_index, neighbor_pixelization_1d_index
                ]

                separation_from_neighbor = (
                    grid[sub_mask_1d_index, 0] - pixel_centres[neighbor, 0]
                ) ** 2 + (grid[sub_mask_1d_index, 1] - pixel_centres[neighbor, 1]) ** 2

                if (
                    separation_from_neighbor
                    < closest_separation_from_pixelization_to_neighbor
                ):
                    closest_separation_from_pixelization_to_neighbor = (
                        separation_from_neighbor
                    )
                    closest_neighbor_pixelization_1d_index = (
                        neighbor_pixelization_1d_index
                    )

            neighboring_pixelization_1d_index = pixel_neighbors[
                nearest_pixelization_1d_index, closest_neighbor_pixelization_1d_index
            ]
            sub_pixel_to_neighboring_pixelization_distance = (
                closest_separation_from_pixelization_to_neighbor
            )

            if (
                sub_pixel_to_nearest_pixelization_distance
                <= sub_pixel_to_neighboring_pixelization_distance
            ):
                sub_mask_1d_index_to_pixeliztion_1d_index[
                    sub_mask_1d_index
                ] = nearest_pixelization_1d_index
                break
            else:
                nearest_pixelization_1d_index = neighboring_pixelization_1d_index

    return sub_mask_1d_index_to_pixeliztion_1d_index



import numpy as np
from astropy import cosmology as cosmo
from skimage import measure

import autofit as af
from autolens import exc, dimensions as dim
from autolens.array import scaled_array
from autolens.array.grids import (
    reshape_array_from_grid,
    reshape_array_from_grid_and_convolver,
    reshape_array_from_grid_and_psf,
    reshape_returned_grid,
)
from autolens.lens.util import lens_util
from autolens.model import cosmology_util


class AbstractPlane(object):
    def __init__(self, redshift, galaxies, cosmology):
        """A plane of galaxies where all galaxies are at the same redshift.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of galaxies in this plane.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        if redshift is None:

            if not galaxies:
                raise exc.RayTracingException(
                    "A redshift and no galaxies were input to a Plane. A redshift for the Plane therefore cannot be"
                    "determined"
                )
            elif not all(
                [galaxies[0].redshift == galaxy.redshift for galaxy in galaxies]
            ):
                raise exc.RayTracingException(
                    "A redshift and two or more galaxies with different redshifts were input to a Plane. A unique "
                    "Redshift for the Plane therefore cannot be determined"
                )
            else:
                redshift = galaxies[0].redshift

        self.redshift = redshift
        self.galaxies = galaxies
        self.cosmology = cosmology

    @property
    def galaxy_redshifts(self):
        return [galaxy.redshift for galaxy in self.galaxies]

    @property
    def has_light_profile(self):
        return any(list(map(lambda galaxy: galaxy.has_light_profile, self.galaxies)))

    @property
    def has_mass_profile(self):
        return any(list(map(lambda galaxy: galaxy.has_mass_profile, self.galaxies)))

    @property
    def has_pixelization(self):
        return any([galaxy.pixelization for galaxy in self.galaxies])

    @property
    def has_regularization(self):
        return any([galaxy.regularization for galaxy in self.galaxies])

    @property
    def galaxies_with_pixelization(self):
        return list(filter(lambda galaxy: galaxy.has_pixelization, self.galaxies))

    @property
    def galaxies_with_regularization(self):
        return list(filter(lambda galaxy: galaxy.has_regularization, self.galaxies))

    @property
    def pixelization(self):

        if len(self.galaxies_with_pixelization) == 0:
            return None
        if len(self.galaxies_with_pixelization) == 1:
            return self.galaxies_with_pixelization[0].pixelization
        elif len(self.galaxies_with_pixelization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with pixelizations in one plane is above 1"
            )

    @property
    def regularization(self):

        if len(self.galaxies_with_regularization) == 0:
            return None
        if len(self.galaxies_with_regularization) == 1:
            return self.galaxies_with_regularization[0].regularization
        elif len(self.galaxies_with_regularization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with regularizations in one plane is above 1"
            )

    @property
    def binned_hyper_galaxy_image_1d_of_galaxy_with_pixelization(self):
        galaxies_with_pixelization = self.galaxies_with_pixelization
        if galaxies_with_pixelization:
            return galaxies_with_pixelization[0].binned_hyper_galaxy_image_1d

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda galaxy: galaxy.has_hyper_galaxy, self.galaxies)))

    @property
    def centres_of_galaxy_mass_profiles(self):

        galaxies_with_mass_profiles = [
            galaxy for galaxy in self.galaxies if galaxy.has_mass_profile
        ]

        mass_profile_centres = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_centres[galaxy_index] = [
                profile.centre for profile in galaxy.mass_profiles
            ]
        return mass_profile_centres

    @property
    def axis_ratios_of_galaxy_mass_profiles(self):
        galaxies_with_mass_profiles = [
            galaxy for galaxy in self.galaxies if galaxy.has_mass_profile
        ]

        mass_profile_axis_ratios = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_axis_ratios[galaxy_index] = [
                profile.axis_ratio for profile in galaxy.mass_profiles
            ]
        return mass_profile_axis_ratios

    @property
    def phis_of_galaxy_mass_profiles(self):

        galaxies_with_mass_profiles = [
            galaxy for galaxy in self.galaxies if galaxy.has_mass_profile
        ]

        mass_profile_phis = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_phis[galaxy_index] = [
                profile.phi for profile in galaxy.mass_profiles
            ]
        return mass_profile_phis


class AbstractPlaneCosmology(AbstractPlane):
    def __init__(self, redshift, galaxies, cosmology):

        super(AbstractPlaneCosmology, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    @property
    def arcsec_per_kpc(self):
        return cosmology_util.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=self.redshift, cosmology=self.cosmology
        )

    @property
    def kpc_per_arcsec(self):
        return 1.0 / self.arcsec_per_kpc

    def angular_diameter_distance_to_earth_in_units(self, unit_length="arcsec"):
        return cosmology_util.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=self.redshift, cosmology=self.cosmology, unit_length=unit_length
        )

    def cosmic_average_density_in_units(
        self, unit_length="arcsec", unit_mass="solMass"
    ):
        return cosmology_util.cosmic_average_density_from_redshift_and_cosmology(
            redshift=self.redshift,
            cosmology=self.cosmology,
            unit_length=unit_length,
            unit_mass=unit_mass,
        )


class AbstractPlaneLensing(AbstractPlaneCosmology):
    def __init__(self, redshift, galaxies, cosmology):
        super(AbstractPlaneCosmology, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    @reshape_array_from_grid
    def profile_image_from_grid(self, grid, return_in_2d=True, return_binned=True):
        """Compute the profile-image plane image of the list of galaxies of the plane's sub-grid, by summing the
        individual images of each galaxy's light profile.

        The image is calculated on the sub-grid and binned-up to the original grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an array of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.galaxies:
            return sum(
                map(
                    lambda galaxy: galaxy.profile_image_from_grid(
                        grid=grid, return_in_2d=False, return_binned=False
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((grid.shape[0]), 0.0)

    def profile_images_of_galaxies_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):
        return list(
            map(
                lambda galaxy: self.profile_image_of_galaxy_from_grid_and_galaxy(
                    grid=grid,
                    galaxy=galaxy,
                    return_in_2d=return_in_2d,
                    return_binned=return_binned,
                ),
                self.galaxies,
            )
        )

    def profile_image_of_galaxy_from_grid_and_galaxy(
        self, grid, galaxy, return_in_2d=True, return_binned=True
    ):
        return galaxy.profile_image_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

    @reshape_array_from_grid
    def convergence_from_grid(self, grid, return_in_2d=True, return_binned=True):
        """Compute the convergence of the list of galaxies of the plane's sub-grid, by summing the individual convergences \
        of each galaxy's mass profile.

        The convergence is calculated on the sub-grid and binned-up to the original grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an array of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        grid : Grid
            The grid (or sub) of (y,x) arc-second coordinates at the centre of every unmasked pixel which the \
            potential is calculated on.
        galaxies : [galaxy.Galaxy]
            The galaxies whose mass profiles are used to compute the surface densities.
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.convergence_from_grid(
                        grid=grid, return_in_2d=False, return_binned=False
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((grid.shape[0]), 0.0)

    @reshape_array_from_grid
    def potential_from_grid(self, grid, return_in_2d=True, return_binned=True):
        """Compute the potential of the list of galaxies of the plane's sub-grid, by summing the individual potentials \
        of each galaxy's mass profile.

        The potential is calculated on the sub-grid and binned-up to the original grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an array of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        grid : Grid
            The grid (or sub) of (y,x) arc-second coordinates at the centre of every unmasked pixel which the \
            potential is calculated on.
        galaxies : [galaxy.Galaxy]
            The galaxies whose mass profiles are used to compute the surface densities.
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.potential_from_grid(
                        grid=grid, return_in_2d=False, return_binned=False
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((grid.shape[0]), 0.0)

    @reshape_returned_grid
    def deflections_from_grid(self, grid, return_in_2d=True, return_binned=True):
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.deflections_from_grid(
                        grid=grid, return_in_2d=False, return_binned=False
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((grid.shape[0], 2), 0.0)

    @reshape_returned_grid
    def traced_grid_from_grid(self, grid, return_in_2d=True):
        """Trace this plane's grid_stacks to the next plane, using its deflection angles."""

        return grid - self.deflections_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

    @reshape_returned_grid
    def deflections_via_potential_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):
        potential_2d = self.potential_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        deflections_y_2d = np.gradient(potential_2d, grid.in_2d[:, 0, 0], axis=0)
        deflections_x_2d = np.gradient(potential_2d, grid.in_2d[0, :, 1], axis=1)

        return np.stack((deflections_y_2d, deflections_x_2d), axis=-1)

    @reshape_array_from_grid
    def lensing_jacobian_a11_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1.0 - np.gradient(deflections_2d[:, :, 1], grid.in_2d[0, :, 1], axis=1)

    @reshape_array_from_grid
    def lensing_jacobian_a12_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(deflections_2d[:, :, 1], grid.in_2d[:, 0, 0], axis=0)

    @reshape_array_from_grid
    def lensing_jacobian_a21_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(deflections_2d[:, :, 0], grid.in_2d[0, :, 1], axis=1)

    @reshape_array_from_grid
    def lensing_jacobian_a22_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1 - np.gradient(deflections_2d[:, :, 0], grid.in_2d[:, 0, 0], axis=0)

    def lensing_jacobian_from_grid(self, grid, return_in_2d=True, return_binned=True):

        a11 = self.lensing_jacobian_a11_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a12 = self.lensing_jacobian_a12_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a21 = self.lensing_jacobian_a21_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a22 = self.lensing_jacobian_a22_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        return np.array([[a11, a12], [a21, a22]])

    @reshape_array_from_grid
    def convergence_via_jacobian_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        convergence = 1 - 0.5 * (jacobian[0, 0] + jacobian[1, 1])

        return convergence

    @reshape_array_from_grid
    def shear_via_jacobian_from_grid(self, grid, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        gamma_1 = 0.5 * (jacobian[1, 1] - jacobian[0, 0])
        gamma_2 = -0.5 * (jacobian[0, 1] + jacobian[1, 0])

        return (gamma_1 ** 2 + gamma_2 ** 2) ** 0.5

    @reshape_array_from_grid
    def tangential_eigen_value_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        convergence = self.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        shear = self.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        return 1 - convergence - shear

    @reshape_array_from_grid
    def radial_eigen_value_from_grid(self, grid, return_in_2d=True, return_binned=True):

        convergence = self.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        shear = self.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        return 1 - convergence + shear

    @reshape_array_from_grid
    def magnification_from_grid(self, grid, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        det_jacobian = jacobian[0, 0] * jacobian[1, 1] - jacobian[0, 1] * jacobian[1, 0]

        return 1 / det_jacobian

    def tangential_critical_curve_from_grid(self, grid):

        lambda_tangential_2d = self.tangential_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        tangential_critical_curve_indices = measure.find_contours(
            lambda_tangential_2d, 0
        )

        if tangential_critical_curve_indices == []:
            return []

        return grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=tangential_critical_curve_indices[0],
            shape=lambda_tangential_2d.shape,
        )

    def radial_critical_curve_from_grid(self, grid):

        lambda_radial_2d = self.radial_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        radial_critical_curve_indices = measure.find_contours(lambda_radial_2d, 0)

        if radial_critical_curve_indices == []:
            return []

        return grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=radial_critical_curve_indices[0], shape=lambda_radial_2d.shape
        )

    def tangential_caustic_from_grid(self, grid):

        tangential_critical_curve = self.tangential_critical_curve_from_grid(grid=grid)

        if tangential_critical_curve == []:
            return []

        deflections_1d = self.deflections_from_grid(
            grid=tangential_critical_curve, return_in_2d=False, return_binned=False
        )

        return tangential_critical_curve - deflections_1d

    def radial_caustic_from_grid(self, grid):

        radial_critical_curve = self.radial_critical_curve_from_grid(grid=grid)

        if radial_critical_curve == []:
            return []

        deflections_1d = self.deflections_from_grid(
            grid=radial_critical_curve, return_in_2d=False, return_binned=False
        )

        return radial_critical_curve - deflections_1d

    def critical_curves_from_grid(self, grid):
        return [
            self.tangential_critical_curve_from_grid(grid=grid),
            self.radial_critical_curve_from_grid(grid=grid),
        ]

    def caustics_from_grid(self, grid):
        return [
            self.tangential_caustic_from_grid(grid=grid),
            self.radial_caustic_from_grid(grid=grid),
        ]

    def luminosities_of_galaxies_within_circles_in_units(
        self, radius: dim.Length, unit_luminosity="eps", exposure_time=None
    ):
        """Compute the total luminosity of all galaxies in this plane within a circle of specified radius.

        See *galaxy.light_within_circle* and *light_profiles.light_within_circle* for details \
        of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second units to counts.
        """
        return list(
            map(
                lambda galaxy: galaxy.luminosity_within_circle_in_units(
                    radius=radius,
                    unit_luminosity=unit_luminosity,
                    exposure_time=exposure_time,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def luminosities_of_galaxies_within_ellipses_in_units(
        self, major_axis: dim.Length, unit_luminosity="eps", exposure_time=None
    ):
        """
        Compute the total luminosity of all galaxies in this plane within a ellipse of specified major-axis.

        The value returned by this integral is dimensionless, and a conversion factor can be specified to convert it \
        to a physical value (e.g. the photometric zeropoint).

        See *galaxy.light_within_ellipse* and *light_profiles.light_within_ellipse* for details
        of how this is performed.

        Parameters
        ----------
        major_axis : float
            The major-axis radius of the ellipse.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second units to counts.
        """
        return list(
            map(
                lambda galaxy: galaxy.luminosity_within_ellipse_in_units(
                    major_axis=major_axis,
                    unit_luminosity=unit_luminosity,
                    exposure_time=exposure_time,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def masses_of_galaxies_within_circles_in_units(
        self, radius: dim.Length, unit_mass="solMass", redshift_source=None
    ):
        """Compute the total mass of all galaxies in this plane within a circle of specified radius.

        See *galaxy.angular_mass_within_circle* and *mass_profiles.angular_mass_within_circle* for details
        of how this is performed.

        Parameters
        ----------
        redshift_source
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_mass : str
            The units the mass is returned in (angular | solMass).

        """
        return list(
            map(
                lambda galaxy: galaxy.mass_within_circle_in_units(
                    radius=radius,
                    unit_mass=unit_mass,
                    redshift_source=redshift_source,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def masses_of_galaxies_within_ellipses_in_units(
        self, major_axis: dim.Length, unit_mass="solMass", redshift_source=None
    ):
        """Compute the total mass of all galaxies in this plane within a ellipse of specified major-axis.

        See *galaxy.angular_mass_within_ellipse* and *mass_profiles.angular_mass_within_ellipse* for details \
        of how this is performed.

        Parameters
        ----------
        redshift_source
        unit_mass
        major_axis : float
            The major-axis radius of the ellipse.

        """
        return list(
            map(
                lambda galaxy: galaxy.mass_within_ellipse_in_units(
                    major_axis=major_axis,
                    unit_mass=unit_mass,
                    redshift_source=redshift_source,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def einstein_radius_in_units(self, unit_length="arcsec"):

        if self.has_mass_profile:
            return sum(
                filter(
                    None,
                    list(
                        map(
                            lambda galaxy: galaxy.einstein_radius_in_units(
                                unit_length=unit_length, cosmology=self.cosmology
                            ),
                            self.galaxies,
                        )
                    ),
                )
            )

    def einstein_mass_in_units(self, unit_mass="solMass", redshift_source=None):

        if self.has_mass_profile:
            return sum(
                filter(
                    None,
                    list(
                        map(
                            lambda galaxy: galaxy.einstein_mass_in_units(
                                unit_mass=unit_mass,
                                redshift_source=redshift_source,
                                cosmology=self.cosmology,
                            ),
                            self.galaxies,
                        )
                    ),
                )
            )


class AbstractPlaneData(AbstractPlaneLensing):
    def __init__(self, redshift, galaxies, cosmology):

        super(AbstractPlaneData, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    @reshape_array_from_grid_and_psf
    def blurred_profile_image_from_grid_and_psf(
        self, grid, psf, preload_blurring_grid=None, return_in_2d=True
    ):

        profile_image = self.profile_image_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        if preload_blurring_grid is None:
            preload_blurring_grid = grid.blurring_grid_from_psf_shape(
                psf_shape=psf.shape
            )

        blurring_image = self.profile_image_from_grid(
            grid=preload_blurring_grid, return_in_2d=True, return_binned=True
        )

        return psf.convolve(profile_image + blurring_image)

    def blurred_profile_images_of_galaxies_from_grid_and_psf(
        self, grid, psf, preload_blurring_grid=None, return_in_2d=True
    ):
        return [
            galaxy.blurred_profile_image_from_grid_and_psf(
                grid=grid,
                psf=psf,
                preload_blurring_grid=preload_blurring_grid,
                return_in_2d=return_in_2d,
            )
            for galaxy in self.galaxies
        ]

    @reshape_array_from_grid_and_convolver
    def blurred_profile_image_from_grid_and_convolver(
        self, grid, convolver, preload_blurring_grid=None, return_in_2d=True
    ):

        if preload_blurring_grid is None:
            preload_blurring_grid = grid.blurring_grid_from_psf_shape(
                psf_shape=convolver.psf.shape
            )

        if convolver.blurring_mask is None:
            blurring_mask = grid.mask.blurring_mask_from_psf_shape(
                psf_shape=convolver.psf.shape
            )
            convolver = convolver.convolver_with_blurring_mask_added(
                blurring_mask=blurring_mask
            )

        image_array = self.profile_image_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        blurring_array = self.profile_image_from_grid(
            grid=preload_blurring_grid, return_in_2d=False, return_binned=True
        )

        return convolver.convolve_image(
            image_array=image_array, blurring_array=blurring_array
        )

    def blurred_profile_images_of_galaxies_from_grid_and_convolver(
        self, grid, convolver, preload_blurring_grid=None, return_in_2d=True
    ):
        return [
            galaxy.blurred_profile_image_from_grid_and_convolver(
                grid=grid,
                convolver=convolver,
                preload_blurring_grid=preload_blurring_grid,
                return_in_2d=return_in_2d,
            )
            for galaxy in self.galaxies
        ]

    def visibilities_from_grid_and_transformer(self, grid, transformer):

        profile_image_plane_image_1d = self.profile_image_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        return transformer.visibilities_from_image_1d(
            image_1d=profile_image_plane_image_1d
        )

    def pixelization_grid_from_grid(self, grid):

        if not self.has_pixelization:
            return None

        binned_hyper_galaxy_image_1d = (
            self.binned_hyper_galaxy_image_1d_of_galaxy_with_pixelization
        )

        return self.pixelization.pixelization_grid_from_grid(
            grid=grid,
            cluster_grid=grid.binned,
            hyper_image=binned_hyper_galaxy_image_1d,
        )

    def mapper_from_grid_and_pixelization_grid(
        self, grid, pixelization_grid, inversion_uses_border=False
    ):

        galaxies_with_pixelization = list(
            filter(lambda galaxy: galaxy.pixelization is not None, self.galaxies)
        )

        if len(galaxies_with_pixelization) == 0:
            return None
        if len(galaxies_with_pixelization) == 1:

            pixelization = galaxies_with_pixelization[0].pixelization

            return pixelization.mapper_from_grid_and_pixelization_grid(
                grid=grid,
                pixelization_grid=pixelization_grid,
                inversion_uses_border=inversion_uses_border,
                hyper_image=galaxies_with_pixelization[0].hyper_galaxy_image_1d,
            )

        elif len(galaxies_with_pixelization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with pixelizations in one plane is above 1"
            )

    def plane_image_from_grid(self, grid):
        return lens_util.plane_image_of_galaxies_from_grid(
            shape=grid.mask.shape,
            grid=grid.unlensed_unsubbed_1d,
            galaxies=self.galaxies,
        )

    def hyper_noise_map_1d_from_noise_map_1d(self, noise_map_1d):
        hyper_noise_maps_1d = self.hyper_noise_maps_1d_of_galaxies_from_noise_map_1d(
            noise_map_1d=noise_map_1d
        )
        hyper_noise_maps_1d = [
            hyper_noise_map
            for hyper_noise_map in hyper_noise_maps_1d
            if hyper_noise_map is not None
        ]
        return sum(hyper_noise_maps_1d)

    def hyper_noise_maps_1d_of_galaxies_from_noise_map_1d(self, noise_map_1d):
        """For a contribution map and noise-map, use the model hyper_galaxy galaxies to compute a hyper noise-map.

        Parameters
        -----------
        noise_map_1d : ccd.NoiseMap or ndarray
            An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        """
        hyper_noise_maps_1d = []

        for galaxy in self.galaxies:
            if galaxy.hyper_galaxy is not None:

                hyper_noise_map_1d = galaxy.hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                    noise_map=noise_map_1d,
                    hyper_model_image=galaxy.hyper_model_image_1d,
                    hyper_galaxy_image=galaxy.hyper_galaxy_image_1d,
                )

                hyper_noise_maps_1d.append(hyper_noise_map_1d)

            else:

                hyper_noise_maps_1d.append(None)

        return hyper_noise_maps_1d

    @property
    def contribution_maps_1d_of_galaxies(self):

        contribution_maps_1d = []

        for galaxy in self.galaxies:

            if galaxy.hyper_galaxy is not None:

                contribution_map = galaxy.hyper_galaxy.contribution_map_from_hyper_images(
                    hyper_model_image=galaxy.hyper_model_image_1d,
                    hyper_galaxy_image=galaxy.hyper_galaxy_image_1d,
                )

                contribution_maps_1d.append(contribution_map)

            else:

                contribution_maps_1d.append(None)

        return contribution_maps_1d

    @property
    def yticks(self):
        """Compute the yticks labels of this grid, used for plotting the y-axis ticks when visualizing an image \
        """
        return np.linspace(np.amin(self.grid[:, 0]), np.amax(self.grid[:, 0]), 4)

    @property
    def xticks(self):
        """Compute the xticks labels of this grid, used for plotting the x-axis ticks when visualizing an \
        image"""
        return np.linspace(np.amin(self.grid[:, 1]), np.amax(self.grid[:, 1]), 4)


class Plane(AbstractPlaneData):
    def __init__(self, redshift=None, galaxies=None, cosmology=cosmo.Planck15):

        super(Plane, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

    # noinspection PyUnusedLocal
    def summarize_in_units(
        self,
        radii,
        whitespace=80,
        unit_length="arcsec",
        unit_luminosity="eps",
        unit_mass="solMass",
        redshift_source=None,
        **kwargs
    ):

        summary = ["Plane\n"]
        prefix_plane = ""

        summary += [
            af.text_util.label_and_value_string(
                label=prefix_plane + "redshift",
                value=self.redshift,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        summary += [
            af.text_util.label_and_value_string(
                label=prefix_plane + "kpc_per_arcsec",
                value=self.kpc_per_arcsec,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        angular_diameter_distance_to_earth = self.angular_diameter_distance_to_earth_in_units(
            unit_length=unit_length
        )

        summary += [
            af.text_util.label_and_value_string(
                label=prefix_plane + "angular_diameter_distance_to_earth",
                value=angular_diameter_distance_to_earth,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        for galaxy in self.galaxies:
            summary += ["\n"]
            summary += galaxy.summarize_in_units(
                radii=radii,
                whitespace=whitespace,
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                unit_mass=unit_mass,
                redshift_source=redshift_source,
                cosmology=self.cosmology,
            )

        return summary


class PlanePositions(object):
    def __init__(
        self, redshift, galaxies, positions, compute_deflections=True, cosmology=None
    ):
        """A plane represents a set of galaxies at a given redshift in a ray-tracer_normal and the positions of image-plane \
        coordinates which mappers close to one another in the source-plane.

        Parameters
        -----------
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        positions : [[[]]]
            The (y,x) arc-second coordinates of image-plane pixels which (are expected to) mappers to the same
            location(s) in the final source-plane.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        """

        self.redshift = redshift
        self.galaxies = galaxies
        self.positions = positions

        if compute_deflections:

            def calculate_deflections(pos):
                return sum(
                    map(lambda galaxy: galaxy.deflections_from_grid(pos), galaxies)
                )

            self.deflections = list(
                map(lambda pos: calculate_deflections(pos), self.positions)
            )

        self.cosmology = cosmology

    def trace_to_next_plane(self):
        """Trace the positions to the next plane."""
        return list(
            map(
                lambda positions, deflections: np.subtract(positions, deflections),
                self.positions,
                self.deflections,
            )
        )


class PlaneImage(scaled_array.ScaledRectangularPixelArray):
    def __init__(self, array, pixel_scales, grid, origin=(0.0, 0.0)):
        self.grid = grid
        super(PlaneImage, self).__init__(
            array=array, pixel_scales=pixel_scales, origin=origin
        )



import sys
import numpy as np

sys.path.append("../")

from workspace_jam.plotting.density_plots import density_plot_tools


ltm_skip = 0
center_skip = 0

image_dir = (
    "/gpfs/data/pdtw24/PL_Data/SL03_2/"
)  # Dir of Object to make evidence tables from

image_name = "SLACSJ0252+0039"
image_name = "SLACSJ1250+0523"
# image_name = 'SLACSJ1430+4105'

slacs = density_plot_tools.SLACS(image_dir, image_name)

# pipeline_folder = 'Pipeline_Total'
# phase_folder = 'PL_Phase_8'

# pipeline_folder = 'Pipeline_LMDM'
# phase_folder = 'PL_Phase_7'

pipeline_folder = "Pipeline_LTM2"
phase_folder = "PL_Phase_7"

if image_name == "SLACSJ1250+0523" or image_name == "SLACSJ1430+4105":
    center_skip = 2

if pipeline_folder == "Pipeline_LTM2":
    ltm_skip = 1

image_dir = image_dir + image_name + "/" + pipeline_folder + "/" + phase_folder + "/"

model_folder = density_plot_tools.getModelFolders(image_dir)

pdf_file = image_dir + model_folder[0] + "/" + image_name + ".txt"

slacs.load_samples(pdf_file, center_skip, ltm_skip)

model_indexes, sample_weights, total_masses, stellar_masses, bulge_masses, disk_masses, dark_masses, stellar_fractions = slacs.masses_of_all_samples(
    radius_kpc=10.0
)

total_mass, total_mass_std = density_plot_tools.weighted_avg_and_std(
    total_masses, sample_weights
)
total_mass_lower = total_mass - total_mass_std
total_mass_upper = total_mass + total_mass_std

stellar_mass, stellar_mass_std = density_plot_tools.weighted_avg_and_std(
    stellar_masses, sample_weights
)
stellar_mass_lower = stellar_mass - stellar_mass_std
stellar_mass_upper = stellar_mass + stellar_mass_std

bulge_mass, bulge_mass_std = density_plot_tools.weighted_avg_and_std(
    bulge_masses, sample_weights
)
bulge_mass_lower = bulge_mass - bulge_mass_std
bulge_mass_upper = bulge_mass + bulge_mass_std

disk_mass, disk_mass_std = density_plot_tools.weighted_avg_and_std(
    disk_masses, sample_weights
)
disk_mass_lower = disk_mass - disk_mass_std
disk_mass_upper = disk_mass + disk_mass_std

dark_mass, dark_mass_std = density_plot_tools.weighted_avg_and_std(
    dark_masses, sample_weights
)
dark_mass_lower = dark_mass - dark_mass_std
dark_mass_upper = dark_mass + dark_mass_std

print()
print(
    "Total Mass = ",
    "{0:e}".format(total_mass),
    "(",
    "{0:e}".format(total_mass_std),
    "{0:e}".format(total_mass_upper),
    ")",
)
print(
    "stellar Mass = ",
    "{0:e}".format(stellar_mass),
    "(",
    "{0:e}".format(stellar_mass_std),
    "{0:e}".format(stellar_mass_upper),
    ")",
)
print(
    "bulge Mass = ",
    "{0:e}".format(bulge_mass),
    "(",
    "{0:e}".format(bulge_mass_std),
    "{0:e}".format(bulge_mass_upper),
    ")",
)
print(
    "disk Mass = ",
    "{0:e}".format(disk_mass),
    "(",
    "{0:e}".format(disk_mass_std),
    "{0:e}".format(disk_mass_upper),
    ")",
)
print(
    "dark Mass = ",
    "{0:e}".format(dark_mass),
    "(",
    "{0:e}".format(dark_mass_std),
    "{0:e}".format(dark_mass_upper),
    ")",
)
print("SLACS Mass in R_Ein= ", slacs.mass)
print("Our Mass / SLACS Mass = ", total_mass / float(slacs.mass))
#
# print()
# print('Stellar Mass fraction in R_Ein = ', stellar_mass / total_mass, '(', stellar_mass_lower / total_mass, stellar_mass_upper / total_mass, ')' )
# print('SLACS Chabrier Stellar Mass Fraction in R_Ein = ', slacs.chab_stellar_frac, ' +- ', slacs.chab_stellar_frac_error )
# print('SLACS Salpeter Stellar Mass Fraction in R_Ein = ', slacs.sal_stellar_frac, ' +- ', slacs.sal_stellar_frac_error  )
#
# print()
# print('Our Stellar Mass = ','{0:e}'.format((stellar_mass)), '(', '{0:e}'.format(stellar_mass_lower), '{0:e}'.format(stellar_mass_upper), ')' )
# print('SLACS Chabrier stellar mass = ', '{0:e}'.format(float(slacs.chab_stellar_mass)), '(', '{0:e}'.format(float(slacs.chab_stellar_mass_lower)), '{0:e}'.format(float(slacs.chab_stellar_mass_upper)), ')'  )
# print('SLACS Salpeter stellar mass = ', '{0:e}'.format(float(slacs.sal_stellar_mass)), '(', '{0:e}'.format(float(slacs.sal_stellar_mass_lower)), '{0:e}'.format(float(slacs.sal_stellar_mass_upper)), ')'  )
# print('Our Stellar Mass / SLACS Chabrier Stellar Mass = ', stellar_mass / float(slacs.chab_stellar_mass))
# print('Our Stellar Mass / SLACS Salpeter Stellar Mass = ', stellar_mass / float(slacs.sal_stellar_mass))


slacs.weighted_densities_vs_radii(radius_kpc=15.0, weight_cut=1e-4, number_bins=50)

slacs.plot_density(image_name=image_name, labels=["Sersic", "Exponential", "NFWSph"])
